from .ldata import bitPosArray, fileBits, rankBits, moveArray
from .bitboard import firstBit, iterBits
from .validator import validateMove

from pychess.Utils.const import (
    SAN,
    AN,
    LAN,
    ENPASSANT,
    EMPTY,
    PAWN,
    KING_CASTLE,
    QUEEN_CASTLE,
    reprFile,
    reprRank,
    chr2Sign,
    cordDic,
    reprSign,
    reprCord,
    reprSignSittuyin,
    reprSignMakruk,
    QUEEN,
    KNIGHT,
    BISHOP,
    ROOK,
    KING,
    NORMALCHESS,
    NORMAL_MOVE,
    PROMOTIONS,
    WHITE,
    BLACK,
    DROP,
    FAN_PIECES,
    SITTUYINCHESS,
    FISCHERRANDOMCHESS,
    SUICIDECHESS,
    MAKRUKCHESS,
    CAMBODIANCHESS,
    GIVEAWAYCHESS,
    ATOMICCHESS,
    WILDCASTLECHESS,
    WILDCASTLESHUFFLECHESS,
    HORDECHESS,
    SCHESS,
    LIGHTBRIGADECHESS,
    chrU2Sign,
    CASTLE_KR,
    CASTLE_SAN,
    QUEEN_PROMOTION,
    KNIGHT_PROMOTION,
    FAN,
    ASEAN_QUEEN,
    HAWK_PROMOTION,
    HAWK_GATE,
    ELEPHANT_GATE,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
    GATINGS,
)
from pychess.Utils.repr import reprPiece, localReprSign
from pychess.Utils.lutils.lmovegen import (
    genAllMoves,
    genPieceMoves,
    newMove,
    gen_sittuyin_promotions,
)


def RANK(cord):
    return cord >> 3


def FILE(cord):
    return cord & 7


def TCORD(move):
    return move & 63


def FCORD(move):
    return move >> 6 & 63


def FLAG(move):
    return move >> 12


def PROMOTE_PIECE(flag):
    return flag - 2


def GATE_PIECE(flag):
    return flag - 6 if flag >= HAWK_GATE_AT_ROOK else flag - 4


def FLAG_PIECE(piece):
    return piece + 2


class ParsingError(Exception):
    """Please raise this with a 3-tupple: (move, reason, board.asFen())
    The reason should be usable in the context: 'Move was not parseable
    because %s' % reason"""

    pass


def sittuyin_promotion_fcord(board, tcord):
    valid_promotions = list(gen_sittuyin_promotions(board))
    queenMoves = moveArray[ASEAN_QUEEN]
    for fcord in iterBits(queenMoves[tcord]):
        if board.arBoard[fcord] and board.arBoard[fcord] == PAWN:
            if newMove(fcord, tcord, QUEEN_PROMOTION) in valid_promotions:
                return fcord


################################################################################
# parseAny                                                                     #
################################################################################


def parseAny(board, algnot):
    type = determineAlgebraicNotation(algnot)
    if type == SAN:
        return parseSAN(board, algnot)
    if type == AN:
        return parseAN(board, algnot)
    if type == LAN:
        return parseLAN(board, algnot)
    return parseFAN(board, algnot)


def determineAlgebraicNotation(algnot):
    upnot = algnot.upper()
    if upnot in ("O-O", "O-O-O", "0-0", "0-0-0", "OO", "OOO", "00", "000"):
        return SAN

    # Test for e2-e4
    if "-" in algnot:
        return LAN

    # Test for b4xc5
    if "x" in algnot and algnot.split("x")[0] in cordDic:
        return LAN

    # Test for e2e4 or a7a8q or a7a8=q
    if algnot[:2] in cordDic and algnot[2:4] in cordDic:
        return AN

    if algnot[0] in FAN_PIECES[WHITE] or algnot[0] in FAN_PIECES[BLACK]:
        return FAN

    return SAN


################################################################################
# listToSan                                                                    #
################################################################################


def listToSan(board, moves):
    # Work on a copy to ensure we don't break things
    board = board.clone()
    sanmoves = []
    for move in moves:
        san = toSAN(board, move)
        sanmoves.append(san)
        board.applyMove(move)
    return sanmoves


################################################################################
# listToMoves                                                                  #
################################################################################


def listToMoves(board, movstrs, type=None, testvalidate=False, ignoreErrors=False):
    # Work on a copy to ensure we don't break things
    board = board.clone()
    moves = []

    for mstr in movstrs:
        try:
            if type is None:
                move = parseAny(board, mstr)
            elif type == SAN:
                move = parseSAN(board, mstr)
            elif type == AN:
                move = parseAN(board, mstr)
            elif type == LAN:
                move = parseLAN(board, mstr)
        except ParsingError:
            if ignoreErrors:
                break
            raise

        if testvalidate and mstr != "--":
            if not validateMove(board, move):
                if not ignoreErrors:
                    raise ParsingError(mstr, "Validation", board.asFen())
                break

        moves.append(move)
        board.applyMove(move)

    return moves


################################################################################
# toSan                                                                        #
################################################################################


def toSAN(board, move, localRepr=False):
    """Returns a Short/Abbreviated Algebraic Notation string of a move
    The board should be prior to the move"""

    def check_or_mate():
        board_clone = board.clone()
        board_clone.applyMove(move)
        sign = ""
        if board_clone.isChecked():
            for altmove in genAllMoves(board_clone):
                if board.variant == ATOMICCHESS:
                    from pychess.Variants.atomic import kingExplode

                    if kingExplode(
                        board_clone, altmove, 1 - board_clone.color
                    ) and not kingExplode(board_clone, altmove, board_clone.color):
                        sign = "+"
                        break
                    elif kingExplode(board_clone, altmove, board_clone.color):
                        continue
                board_clone.applyMove(altmove)
                if board_clone.opIsChecked():
                    board_clone.popMove()
                    continue
                sign = "+"
                break
            else:
                sign = "#"
        return sign

    flag = move >> 12
    fcord = (move >> 6) & 63

    if flag == KING_CASTLE:
        return "O-O%s" % check_or_mate()
    elif flag == QUEEN_CASTLE:
        return "O-O-O%s" % check_or_mate()

    tcord = move & 63

    # Sittuyin in place promotion is NOT null move
    if fcord == tcord and flag != DROP and flag != QUEEN_PROMOTION:
        return "--"

    fpiece = fcord if flag == DROP else board.arBoard[fcord]
    tpiece = board.arBoard[tcord]

    if flag in GATINGS:
        gate_piece = reprSign[GATE_PIECE(flag)]
        kcastle = (fpiece == KING and fcord - tcord == -2) or (
            flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK) and fcord - tcord > 0
        )
        qcastle = (fpiece == KING and fcord - tcord == 2) or (
            flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK) and fcord - tcord < 0
        )
        if kcastle:
            return f"O-O/{gate_piece}{reprCord[fcord]}{check_or_mate()}"
        elif qcastle:
            return f"O-O-O/{gate_piece}{reprCord[fcord]}{check_or_mate()}"

    part0 = ""
    part1 = ""

    if fpiece != PAWN or flag == DROP:
        if board.variant in (CAMBODIANCHESS, MAKRUKCHESS):
            part0 += reprSignMakruk[fpiece]
        elif board.variant == SITTUYINCHESS:
            part0 += reprSignSittuyin[fpiece]
        elif localRepr:
            part0 += localReprSign[fpiece]
        else:
            part0 += reprSign[fpiece]

    part1 = reprCord[tcord]

    if flag == DROP:
        return f"{part0}@{part1}{check_or_mate()}"

    if fpiece != PAWN or (board.variant == SITTUYINCHESS and flag in PROMOTIONS):
        xs = []
        ys = []

        board_clone = board.clone()
        for altmove in genAllMoves(board_clone, drops=False):
            mfcord = FCORD(altmove)
            if (
                board_clone.arBoard[mfcord] == fpiece
                and mfcord != fcord
                and TCORD(altmove) == tcord
            ):
                board_clone.applyMove(altmove)
                if not board_clone.opIsChecked():
                    xs.append(FILE(mfcord))
                    ys.append(RANK(mfcord))
                board_clone.popMove()

        x = FILE(fcord)
        y = RANK(fcord)

        if ys or xs:
            if y in ys and x not in xs:
                # If we share rank with another piece, but not file
                part0 += reprFile[x]
            elif x in xs and y not in ys:
                # If we share file with another piece, but not rank
                part0 += reprRank[y]
            elif x in xs and y in ys:
                # If we share both file and rank with other pieces
                part0 += reprFile[x] + reprRank[y]
            else:
                # If we doesn't share anything, it is standard to put file
                part0 += reprFile[x]

    if tpiece != EMPTY or flag == ENPASSANT:
        if not (board.variant == SITTUYINCHESS and fcord == tcord):
            part1 = "x" + part1
            if fpiece == PAWN:
                part0 += reprFile[FILE(fcord)]

    notat = part0 + part1
    if flag in PROMOTIONS:
        if board.variant in (CAMBODIANCHESS, MAKRUKCHESS):
            notat += "=" + reprSignMakruk[PROMOTE_PIECE(flag)]
        elif board.variant == SITTUYINCHESS:
            notat += "=" + reprSignSittuyin[PROMOTE_PIECE(flag)]
        elif localRepr:
            notat += "=" + localReprSign[PROMOTE_PIECE(flag)]
        else:
            notat += "=" + reprSign[PROMOTE_PIECE(flag)]

    if flag in GATINGS:
        notat += "/" + reprSign[GATE_PIECE(flag)]

    return f"{notat}{check_or_mate()}"


################################################################################
# parseSan                                                                     #
################################################################################


def parseSAN(board, san):
    """Parse a Short/Abbreviated Algebraic Notation string"""
    notat = san
    color = board.color

    if notat == "--":
        return newMove(board.kings[color], board.kings[color])

    if notat[-1] in "+#":
        notat = notat[:-1]
        # If '++' was used in place of #
        if notat[-1] == "+":
            notat = notat[:-1]

    flag = NORMAL_MOVE

    # If last char is a piece char, we assue it the promote char
    c = notat[-1]
    if c in "KQRBNSMFEHkqrbnsmfeh.":
        c = c.lower()
        if (
            c == "k"
            and board.variant != SUICIDECHESS
            and board.variant != GIVEAWAYCHESS
        ):
            raise ParsingError(san, _("invalid promoted piece"), board.asFen())
        elif c == ".":
            if board.variant in (CAMBODIANCHESS, MAKRUKCHESS, SITTUYINCHESS):
                # temporary hack for xboard bug
                flag = QUEEN_PROMOTION
            else:
                raise ParsingError(san, "invalid san", board.asFen())
        else:
            flag = chr2Sign[c] + 2

        if notat[-2] == "=":
            notat = notat[:-2]
        else:
            notat = notat[:-1]

    if notat[-1] == "/":
        gate_index = san.find("/") + 1
        flag = HAWK_GATE if flag == HAWK_PROMOTION else ELEPHANT_GATE
        notat = notat[:-1]

    if len(notat) < 2:
        raise ParsingError(san, _("the move needs a piece and a cord"), board.asFen())

    if notat[0] in "O0o":
        fcord = board.ini_kings[color]
        if board.variant == SCHESS:
            flag = QUEEN_CASTLE if notat[:5] == "O-O-O" else KING_CASTLE
        else:
            flag = (
                KING_CASTLE
                if notat in ("O-O", "0-0", "o-o", "OO", "00", "oo")
                else QUEEN_CASTLE
            )
        side = flag - QUEEN_CASTLE
        if FILE(fcord) == 3 and board.variant in (
            WILDCASTLECHESS,
            WILDCASTLESHUFFLECHESS,
        ):
            side = 0 if side == 1 else 1
        if board.variant == FISCHERRANDOMCHESS:
            tcord = board.ini_rooks[color][side]
        else:
            tcord = board.fin_kings[color][side]

        if board.variant == SCHESS and "/" in san:
            gate_index = san.find("/") + 1
            if "e" == san[gate_index + 1]:
                flag = HAWK_GATE if san[gate_index] == "H" else ELEPHANT_GATE
            else:
                fcord = board.ini_rooks[color][side]
                tcord = board.ini_kings[color]
                flag = (
                    HAWK_GATE_AT_ROOK
                    if san[gate_index] == "H"
                    else ELEPHANT_GATE_AT_ROOK
                )
            return newMove(fcord, tcord, flag)

        return newMove(fcord, tcord, flag)

    # LAN is not allowed in pgn spec, but sometimes it occures
    if "-" in notat:
        notat = notat.replace("-", "")

    if "@" in notat:
        tcord = cordDic[notat[-2:]]
        if notat[0].islower():
            # Sjeng-ism
            piece = chr2Sign[notat[0]]
        else:
            piece = chrU2Sign[notat[0]]
        return newMove(piece, tcord, DROP)

    # standard piece letters
    if notat[0] in "QRBKNSMFEH":
        piece = chrU2Sign[notat[0]]
        notat = notat[1:]
    # unambigious lowercase piece letters
    elif notat[0] in "qrknsm":
        piece = chr2Sign[notat[0]]
        notat = notat[1:]
    # a lowercase bishop letter or a pawn capture
    elif notat[0] == "b" and len(notat) > 2 and board.variant == NORMALCHESS:
        tcord = cordDic[notat[-2:]]
        trank = int(notat[-1])
        # if from and to lines are not neighbours -> Bishop
        if abs(ord(notat[0]) - ord(notat[-2])) > 1:
            piece = chr2Sign[notat[0]]
            notat = notat[1:]
        # if from and to lines are neighbours (or the same) but to is an empty square
        # which can't be en-passant square target -> Bishop
        elif board.arBoard[tcord] == EMPTY and (
            (color == BLACK and trank != 3) or (color == WHITE and trank != 6)
        ):
            piece = chr2Sign[notat[0]]
            notat = notat[1:]
        # elif "ba3", "bc3" ,"ba6", "bc6"
        # these can be Bishop or Pawn moves, but we don't try to introspect them (sorry)
        else:
            piece = PAWN
    else:
        piece = PAWN
        if notat[-1] in "18" and flag == NORMAL_MOVE and board.variant != SITTUYINCHESS:
            flag = QUEEN_PROMOTION

    if "x" in notat:
        notat, tcord = notat.split("x")
        if tcord not in cordDic:
            raise ParsingError(
                san, _("the captured cord (%s) is incorrect") % tcord, board.asFen()
            )

        tcord = cordDic[tcord]

        if piece == PAWN:
            # If a pawn is attacking an empty cord, we assue it an enpassant
            if board.arBoard[tcord] == EMPTY:
                if (color == BLACK and 2 * 8 <= tcord < 3 * 8) or (
                    color == WHITE and 5 * 8 <= tcord < 6 * 8
                ):
                    flag = ENPASSANT
                else:
                    raise ParsingError(
                        san,
                        _("pawn capture without target piece is invalid"),
                        board.asFen(),
                    )
    else:
        if not notat[-2:] in cordDic:
            raise ParsingError(
                san, _("the end cord (%s) is incorrect") % notat[-2:], board.asFen()
            )

        tcord = cordDic[notat[-2:]]
        notat = notat[:-2]

    # In suicide promoting to king is valid, so
    # more than 1 king per side can exist !
    if (
        board.variant != SUICIDECHESS
        and board.variant != GIVEAWAYCHESS
        and piece == KING
    ):
        return newMove(board.kings[color], tcord, flag)

    # If there is any extra location info, like in the move Bexd1 or Nh3f4 we
    # want to know
    frank = None
    ffile = None
    if notat and notat[0] in reprRank:
        frank = int(notat[0]) - 1
        notat = notat[1:]
    if notat and notat[0] in reprFile:
        ffile = ord(notat[0]) - ord("a")
        notat = notat[1:]
    if notat and notat[0] in reprRank:
        frank = int(notat[0]) - 1
        notat = notat[1:]
        # we know all we want
        return newMove(frank * 8 + ffile, tcord, flag)

    if piece == PAWN:
        if (ffile is not None) and ffile != FILE(tcord):
            # capture
            if color == WHITE:
                fcord = tcord - 7 if ffile > FILE(tcord) else tcord - 9
            else:
                fcord = tcord + 7 if ffile < FILE(tcord) else tcord + 9
        else:
            if color == WHITE:
                pawns = board.boards[WHITE][PAWN]
                # In horde white pawns on first rank may move two squares also
                if (
                    board.variant == HORDECHESS
                    and RANK(tcord) == 2
                    and not (pawns & fileBits[FILE(tcord)] & rankBits[1])
                ):
                    fcord = tcord - 16
                else:
                    fcord = (
                        tcord - 16
                        if RANK(tcord) == 3
                        and not (pawns & fileBits[FILE(tcord)] & rankBits[2])
                        else tcord - 8
                    )
            else:
                pawns = board.boards[BLACK][PAWN]
                fcord = (
                    tcord + 16
                    if RANK(tcord) == 4
                    and not (pawns & fileBits[FILE(tcord)] & rankBits[5])
                    else tcord + 8
                )

            if board.variant == SITTUYINCHESS and flag == QUEEN_PROMOTION:
                if pawns & fileBits[FILE(tcord)] & rankBits[RANK(tcord)]:
                    # in place promotion
                    return newMove(tcord, tcord, flag)
                else:
                    # queen move promotion (fcord have to be the closest cord of promotion zone)
                    fcord = sittuyin_promotion_fcord(board, tcord)
                    return newMove(fcord, tcord, flag)
        return newMove(fcord, tcord, flag)
    else:
        if board.pieceCount[color][piece] == 1:
            # we have only one from this kind if piece, so:
            fcord = firstBit(board.boards[color][piece])
            return newMove(fcord, tcord, flag)
        else:
            # We find all pieces who could have done it. (If san was legal, there should
            # never be more than one)
            moves = genPieceMoves(board, piece, tcord)
            if len(moves) == 1:
                if flag in GATINGS:
                    move = moves.pop()
                    return newMove(FCORD(move), TCORD(move), flag)
                else:
                    return moves.pop()
            else:
                for move in moves:
                    f = FCORD(move)
                    if frank is not None and frank != RANK(f):
                        continue
                    if ffile is not None and ffile != FILE(f):
                        continue
                    board_clone = board.clone()
                    board_clone.applyMove(move)
                    if board_clone.opIsChecked():
                        continue
                    if flag in GATINGS:
                        back_rank = 0 if color == WHITE else 7
                        if RANK(f) == back_rank:
                            return newMove(FCORD(move), TCORD(move), flag)
                        else:
                            continue
                    else:
                        return move

    errstring = _("no %(piece)s is able to move to %(cord)s") % {
        "piece": reprPiece[piece],
        "cord": reprCord[tcord],
    }
    raise ParsingError(san, errstring, board.asFen())


################################################################################
# toLan                                                                        #
################################################################################


def toLAN(board, move, localRepr=False):
    """Returns a Long/Expanded Algebraic Notation string of a move
    board should be prior to the move"""

    fcord = FCORD(move)
    tcord = TCORD(move)
    flag = FLAG(move)
    fpiece = fcord if flag == DROP else board.arBoard[fcord]

    s = ""
    if fpiece != PAWN or flag == DROP:
        if board.variant in (CAMBODIANCHESS, MAKRUKCHESS):
            s = reprSignMakruk[fpiece]
        elif board.variant == SITTUYINCHESS:
            s = reprSignSittuyin[fpiece]
        elif localRepr:
            s = localReprSign[fpiece]
        else:
            s = reprSign[fpiece]

    if flag == DROP:
        s += "@"
    else:
        s += reprCord[FCORD(move)]
        if board.arBoard[tcord] == EMPTY:
            s += "-"
        else:
            s += "x"

    s += reprCord[tcord]

    if flag in PROMOTIONS:
        s += "=" + reprSign[PROMOTE_PIECE(flag)]

    return s


################################################################################
# parseLan                                                                     #
################################################################################


def parseLAN(board, lan):
    """Parse a Long/Expanded Algebraic Notation string"""

    # To parse LAN pawn moves like "e2-e4" as SAN moves, we have to remove a few
    # fields
    if len(lan) == 5:
        if "x" in lan:
            # e4xd5 -> exd5
            return parseSAN(board, lan[0] + lan[3:])
        else:
            # e2-e4 -> e4
            return parseSAN(board, lan[3:])

    # We want to use the SAN parser for LAN moves like "Nb1-c3" or "Rd3xd7"
    # The san parser should be able to handle most stuff, as long as we remove
    # the slash
    if not lan.upper().startswith("O-O") and not lan.startswith("--"):
        lan = lan.replace("-", "")
    return parseSAN(board, lan)


################################################################################
# toAN                                                                         #
################################################################################


def toAN(board, move, short=False, castleNotation=CASTLE_SAN):
    """Returns a Algebraic Notation string of a move
    board should be prior to the move
    short -- returns the short variant, e.g. f7f8q rather than f7f8=Q
    """

    fcord = (move >> 6) & 63
    tcord = move & 63
    flag = move >> 12

    if flag in (KING_CASTLE, QUEEN_CASTLE):
        if castleNotation == CASTLE_SAN:
            return flag == KING_CASTLE and "O-O" or "O-O-O"
        elif castleNotation == CASTLE_KR:
            rooks = board.ini_rooks[board.color]
            tcord = rooks[flag == KING_CASTLE and 1 or 0]
        # No treatment needed for CASTLE_KK

    if flag == DROP:
        if board.variant == SITTUYINCHESS:
            s = f"{reprSignSittuyin[fcord]}@{reprCord[tcord]}"
        else:
            s = f"{reprSign[fcord]}@{reprCord[tcord]}"
    else:
        s = reprCord[fcord] + reprCord[tcord]

    if flag in PROMOTIONS:
        if short:
            if board.variant in (CAMBODIANCHESS, MAKRUKCHESS):
                s += reprSignMakruk[PROMOTE_PIECE(flag)].lower()
            elif board.variant == SITTUYINCHESS:
                s += reprSignSittuyin[PROMOTE_PIECE(flag)].lower()
            else:
                s += reprSign[PROMOTE_PIECE(flag)].lower()
        else:
            if board.variant in (CAMBODIANCHESS, MAKRUKCHESS):
                s += "=" + reprSignMakruk[PROMOTE_PIECE(flag)]
            elif board.variant == SITTUYINCHESS:
                s += "=" + reprSignSittuyin[PROMOTE_PIECE(flag)]
            else:
                s += "=" + reprSign[PROMOTE_PIECE(flag)]

    if flag in GATINGS:
        s += reprSign[GATE_PIECE(flag)].lower()

    return s


################################################################################
# parseAN                                                                      #
################################################################################


def parseAN(board, an):
    """Parse an Algebraic Notation string"""
    length = len(an)
    if not 4 <= length <= 6:
        raise ParsingError(an, "the move must be 4 or 6 chars long", board.asFen())

    if "@" in an:
        tcord = cordDic[an[-2:]]
        if an[0].islower():
            # Sjeng-ism
            piece = chr2Sign[an[0]]
        else:
            piece = chrU2Sign[an[0]]
        return newMove(piece, tcord, DROP)

    try:
        fcord = cordDic[an[:2]]
        tcord = cordDic[an[2:4]]
    except KeyError as e:
        raise ParsingError(an, "the cord (%s) is incorrect" % e.args[0], board.asFen())

    flag = NORMAL_MOVE

    if length > 4 and not an[-1] in "QRBNMSFHEqrbnmsfhe":
        if (
            (board.variant != SUICIDECHESS and board.variant != GIVEAWAYCHESS)
            or (board.variant == SUICIDECHESS or board.variant == GIVEAWAYCHESS)
            and not an[-1] in "Kk"
        ):
            raise ParsingError(an, "invalid promoted piece", board.asFen())

    if length == 5:
        # The a7a8q variant
        if board.variant == SCHESS:
            if an[1] in "18":
                # gating move
                if board.arBoard[fcord] == ROOK and board.arBoard[tcord] == KING:
                    flag = HAWK_GATE_AT_ROOK if an[-1] == "h" else ELEPHANT_GATE_AT_ROOK
                else:
                    flag = HAWK_GATE if an[-1] == "h" else ELEPHANT_GATE
            else:
                # promotion move
                flag = chr2Sign[an[4].lower()] + 2
        else:
            flag = chr2Sign[an[4].lower()] + 2
    elif length == 6:
        # The a7a8=q variant
        flag = chr2Sign[an[5].lower()] + 2
    elif board.arBoard[fcord] == KING:
        if fcord - tcord == 2:
            flag = QUEEN_CASTLE
            if board.variant == FISCHERRANDOMCHESS:
                tcord = board.ini_rooks[board.color][0]
        elif fcord - tcord == -2:
            flag = KING_CASTLE
            if board.variant == FISCHERRANDOMCHESS:
                tcord = board.ini_rooks[board.color][1]
        elif board.arBoard[tcord] == ROOK:
            color = board.color
            friends = board.friends[color]
            if bitPosArray[tcord] & friends:
                if board.ini_rooks[color][0] == tcord:
                    flag = QUEEN_CASTLE
                else:
                    flag = KING_CASTLE
        else:
            flag = NORMAL_MOVE
    elif (
        board.arBoard[fcord] == PAWN
        and board.arBoard[tcord] == EMPTY
        and FILE(fcord) != FILE(tcord)
        and RANK(fcord) != RANK(tcord)
    ):
        flag = ENPASSANT
    elif board.arBoard[fcord] == PAWN:
        # assume queen promotion
        if an[3] in "18":
            if board.variant == LIGHTBRIGADECHESS:
                flag = QUEEN_PROMOTION if board.color == WHITE else KNIGHT_PROMOTION
            elif board.variant != SITTUYINCHESS:
                flag = QUEEN_PROMOTION

    return newMove(fcord, tcord, flag)


################################################################################
# toFAN                                                                        #
################################################################################

san2WhiteFanDic = {
    ord("K"): FAN_PIECES[WHITE][KING],
    ord("Q"): FAN_PIECES[WHITE][QUEEN],
    ord("M"): FAN_PIECES[WHITE][QUEEN],
    ord("F"): FAN_PIECES[WHITE][QUEEN],
    ord("R"): FAN_PIECES[WHITE][ROOK],
    ord("B"): FAN_PIECES[WHITE][BISHOP],
    ord("S"): FAN_PIECES[WHITE][BISHOP],
    ord("N"): FAN_PIECES[WHITE][KNIGHT],
    ord("P"): FAN_PIECES[WHITE][PAWN],
    ord("+"): "†",
    ord("#"): "‡",
}

san2BlackFanDic = {
    ord("K"): FAN_PIECES[BLACK][KING],
    ord("Q"): FAN_PIECES[BLACK][QUEEN],
    ord("M"): FAN_PIECES[BLACK][QUEEN],
    ord("F"): FAN_PIECES[BLACK][QUEEN],
    ord("R"): FAN_PIECES[BLACK][ROOK],
    ord("B"): FAN_PIECES[BLACK][BISHOP],
    ord("S"): FAN_PIECES[BLACK][BISHOP],
    ord("N"): FAN_PIECES[BLACK][KNIGHT],
    ord("P"): FAN_PIECES[BLACK][PAWN],
    ord("+"): "†",
    ord("#"): "‡",
}


def toFAN(board, move):
    """Returns a Figurine Algebraic Notation string of a move"""

    san = toSAN(board, move)
    return san.translate(san2WhiteFanDic)


################################################################################
# parseFAN                                                                     #
################################################################################

fan2SanDic = {}
for k, v in san2WhiteFanDic.items():
    fan2SanDic[ord(v)] = chr(k)
for k, v in san2BlackFanDic.items():
    fan2SanDic[ord(v)] = chr(k)


def parseFAN(board, fan):
    """Parse a Figurine Algebraic Notation string"""

    san = fan.translate(fan2SanDic)
    return parseSAN(board, san)


################################################################################
# toPolyglot                                                                   #
################################################################################


def toPolyglot(board, move):
    """Returns a 16-bit Polyglot-format move
    board should be prior to the move
    """
    pg = move & 4095
    if FLAG(move) in PROMOTIONS:
        pg |= (PROMOTE_PIECE(FLAG(move)) - 1) << 12
    elif FLAG(move) == QUEEN_CASTLE:
        pg = (pg & 4032) | board.ini_rooks[board.color][0]
    elif FLAG(move) == KING_CASTLE:
        pg = (pg & 4032) | board.ini_rooks[board.color][1]

    return pg


################################################################################
# parsePolyglot                                                                #
################################################################################


def parsePolyglot(board, pg):
    """Parse a 16-bit Polyglot-format move"""

    tcord = TCORD(pg)
    fcord = FCORD(pg)
    flag = NORMAL_MOVE
    if pg >> 12:
        flag = FLAG_PIECE((pg >> 12) + 1)
    elif board.arBoard[fcord] == KING:
        if board.arBoard[tcord] == ROOK:
            color = board.color
            friends = board.friends[color]
            if bitPosArray[tcord] & friends:
                if board.ini_rooks[color][0] == tcord:
                    flag = QUEEN_CASTLE
                    if board.variant == NORMALCHESS:  # Want e1c1/e8c8
                        tcord += 2
                else:
                    flag = KING_CASTLE
                    if board.variant == NORMALCHESS:  # Want e1g1/e8g8
                        tcord -= 1
    elif (
        board.arBoard[fcord] == PAWN
        and board.arBoard[tcord] == EMPTY
        and FILE(fcord) != FILE(tcord)
        and RANK(fcord) != RANK(tcord)
    ):
        flag = ENPASSANT

    return newMove(fcord, tcord, flag)



import asyncio
import os
import psutil
import signal
import subprocess
import sys
import time

from gi.repository import GObject, GLib

from pychess.System.Log import log
from pychess.Players.ProtocolEngine import TIME_OUT_SECOND


class SubProcess(GObject.GObject):
    __gsignals__ = {
        "line": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "died": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(
        self, path, args=[], warnwords=[], env=None, cwd=".", lowPriority=False
    ):
        GObject.GObject.__init__(self)

        self.path = path
        self.args = args
        self.warnwords = warnwords
        self.env = env or os.environ
        self.cwd = cwd
        self.lowPriority = lowPriority

        self.defname = os.path.split(path)[1]
        self.defname = self.defname[:1].upper() + self.defname[1:].lower()
        cur_time = time.time()
        self.defname = (
            self.defname,
            time.strftime("%H:%m:%%.3f", time.localtime(cur_time)) % (cur_time % 60),
        )
        log.debug(path + " " + " ".join(self.args), extra={"task": self.defname})

        self.argv = [str(u) for u in [self.path] + self.args]
        self.terminated = False

    async def start(self):
        log.debug(
            "SubProcess.start(): create_subprocess_exec...",
            extra={"task": self.defname},
        )
        if sys.platform == "win32":
            # To prevent engines opening console window
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        else:
            startupinfo = None

        create = asyncio.create_subprocess_exec(
            *self.argv,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            startupinfo=startupinfo,
            env=self.env,
            cwd=self.cwd,
        )
        try:
            self.proc = await asyncio.wait_for(create, TIME_OUT_SECOND)
            self.pid = self.proc.pid
            # print(self.pid, self.path)
            if self.lowPriority:
                proc = psutil.Process(self.pid)
                if sys.platform == "win32":
                    niceness = psutil.BELOW_NORMAL_PRIORITY_CLASS
                else:
                    niceness = 15  # The higher, the lower the priority
                if psutil.__version__ >= "2.0.0":
                    proc.nice(niceness)
                else:
                    proc.set_nice(niceness)
            self.read_stdout_task = asyncio.create_task(
                self.read_stdout(self.proc.stdout)
            )
            self.write_task = None
        except asyncio.TimeoutError:
            log.warning("TimeoutError", extra={"task": self.defname})
            raise
        except GLib.GError:
            log.warning("GLib.GError", extra={"task": self.defname})
            raise
        except Exception:
            e = sys.exc_info()[0]
            log.warning("%s" % e, extra={"task": self.defname})
            raise

    def write(self, line):
        self.write_task = asyncio.create_task(self.write_stdin(self.proc.stdin, line))

    async def write_stdin(self, writer, line):
        if self.terminated:
            return
        try:
            log.debug(line, extra={"task": self.defname})
            writer.write(line.encode())
            await writer.drain()
        except BrokenPipeError:
            log.debug(
                "SubProcess.write_stdin(): BrokenPipeError",
                extra={"task": self.defname},
            )
            self.emit("died")
            self.terminate()
        except ConnectionResetError:
            log.debug(
                "SubProcess.write_stdin(): ConnectionResetError",
                extra={"task": self.defname},
            )
            self.emit("died")
            self.terminate()
        except GLib.GError:
            log.debug(
                "SubProcess.write_stdin(): GLib.GError", extra={"task": self.defname}
            )
            self.emit("died")
            self.terminate()

    async def read_stdout(self, reader):
        while True:
            line = await reader.readline()
            if line:
                await asyncio.sleep(0)

                try:
                    line = line.decode().rstrip()
                    # print(line)
                except UnicodeError:
                    # Some engines send author names in different encodinds (f.e. spike)
                    print("UnicodeError while decoding:", line)
                    continue

                if not line:
                    continue

                for word in self.warnwords:
                    if word in line:
                        log.debug(line, extra={"task": self.defname})
                        break
                else:
                    log.debug(line, extra={"task": self.defname})
                self.emit("line", line)
            else:
                self.emit("died")
                break
        self.terminate()

    def terminate(self):
        if self.write_task is not None:
            self.write_task.cancel()
        self.read_stdout_task.cancel()
        try:
            self.proc.terminate()
            log.debug("SubProcess.terminate()", extra={"task": self.defname})
        except ProcessLookupError:
            log.debug(
                "SubProcess.terminate(): ProcessLookupError",
                extra={"task": self.defname},
            )

        self.terminated = True

    def pause(self):
        if sys.platform != "win32":
            try:
                self.proc.send_signal(signal.SIGSTOP)
            except ProcessLookupError:
                log.debug(
                    "SubProcess.pause(): ProcessLookupError",
                    extra={"task": self.defname},
                )

    def resume(self):
        if sys.platform != "win32":
            try:
                self.proc.send_signal(signal.SIGCONT)
            except ProcessLookupError:
                log.debug(
                    "SubProcess.pause(): ProcessLookupError",
                    extra={"task": self.defname},
                )



"""
This module provides some basic functions for accessing pychess datefiles in
system or user space
"""

import os
import sys

################################################################################
# Locate files in system space                                                 #
################################################################################


def _path_is_child_of(parent_path: str, candidate_child_path: str) -> bool:
    real_parent_path = os.path.realpath(parent_path)
    real_candidate_child_path = os.path.realpath(candidate_child_path)
    try:
        return (
            os.path.commonpath([real_parent_path, real_candidate_child_path])
            == real_parent_path
        )
    except ValueError:  # e.g. "Paths don't have the same drive" on Windows
        return False


# Test if we are installed on the system, frozen or are being run from tar/svn
if getattr(sys, "frozen", False):
    _prefix = os.path.join(os.path.dirname(sys.executable), "share", "pychess")
    _installed = True
else:
    home_local = os.path.expanduser("~") + "/.local"
    if _path_is_child_of(sys.prefix, __file__):
        for sub in (
            "share",
            "games",
            "share/games",
            "local/share",
            "local/games",
            "local/share/games",
        ):
            _prefix = os.path.join(sys.prefix, sub, "pychess")
            if os.path.isdir(os.path.join(_prefix, "pieces")):
                _installed = True
                break
        else:
            raise Exception("can't find the pychess data directory")
    elif _path_is_child_of(home_local, __file__):
        _prefix = os.path.join(home_local, "share", "pychess")
        if os.path.isdir(os.path.join(_prefix, "pieces")):
            _installed = True
        else:
            raise Exception("can't find the pychess data directory")
    else:
        _prefix = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../.."))
        _installed = False


def addDataPrefix(subpath):
    return os.path.abspath(os.path.join(_prefix, subpath))


def getDataPrefix():
    return _prefix


def isInstalled():
    return _installed


################################################################################
# Locate files in user space                                                   #
################################################################################


def __get_user_dir(xdg_env_var, fallback_dir_path):
    return os.environ.get(
        xdg_env_var, os.path.join(os.path.expanduser("~"), fallback_dir_path)
    )


def get_user_data_dir():
    return __get_user_dir("XDG_DATA_HOME", ".local/share")


def get_user_config_dir():
    return __get_user_dir("XDG_CONFIG_HOME", ".config")


def get_user_cache_dir():
    return __get_user_dir("XDG_CACHE_HOME", ".cache")


pychess = "pychess"


def getUserDataPrefix():
    return os.path.join(get_user_data_dir(), pychess)


def addUserDataPrefix(subpath):
    return os.path.join(getUserDataPrefix(), subpath)


def getEngineDataPrefix():
    return os.path.join(getUserDataPrefix(), "engines")


def addEngineDataPrefix(subpath):
    return os.path.join(getEngineDataPrefix(), subpath)


def getUserConfigPrefix():
    return os.path.join(get_user_config_dir(), pychess)


def addUserConfigPrefix(subpath):
    return os.path.join(getUserConfigPrefix(), subpath)


def getUserCachePrefix():
    return os.path.join(get_user_cache_dir(), pychess)


def addUserCachePrefix(subpath):
    return os.path.join(getUserCachePrefix(), subpath)


for directory in (
    getUserDataPrefix(),
    getEngineDataPrefix(),
    getUserConfigPrefix(),
    getUserCachePrefix(),
):
    if not os.path.isdir(directory):
        os.makedirs(directory, mode=0o700)



from pychess.Utils.repr import reprColor, reprPiece


class Piece:
    def __init__(self, color, piece, captured=False):
        self.color = color
        self.piece = piece
        self.captured = captured

        # in crazyhouse we need to know this for later captures
        self.promoted = False

        self.opacity = 1.0
        self.x = None
        self.y = None

    # Sign is a deprecated synonym for piece
    def _set_sign(self, sign):
        self.piece = sign

    def _get_sign(self):
        return self.piece

    sign = property(_get_sign, _set_sign)

    def __repr__(self):
        represen = f"<{reprColor[self.color]} {reprPiece[self.piece]}"
        if self.opacity != 1.0:
            represen += " Op:%0.1f" % self.opacity
        if self.x is not None or self.y is not None:
            if self.x is not None:
                represen += " X:%0.1f" % self.x
            else:
                represen += " X:None"
            if self.y is not None:
                represen += " Y:%0.1f" % self.y
            else:
                represen += " Y:None"
        represen += ">"
        return represen



from gi.repository import Gtk, GObject


class BorderBox(Gtk.Alignment):
    def __init__(self, widget=None, top=False, right=False, bottom=False, left=False):
        GObject.GObject.__init__(self)
        self.connect("draw", self._onExpose)

        if widget:
            self.add(widget)

        self.__top = top
        self.__right = right
        self.__bottom = bottom
        self.__left = left
        self._updateBorders()

    def _onExpose(self, area, ctx):
        context = self.get_window().cairo_create()

        style_ctxt = self.get_style_context()
        color = style_ctxt.lookup_color("p_dark_color")[1]
        red, green, blue, alpha = color.red, color.green, color.blue, color.alpha
        context.set_source_rgba(red, green, blue, alpha)

        allocation = self.get_allocation()
        x_loc = allocation.x + 0.5
        y_loc = allocation.y + 0.5
        width = allocation.width - 1
        height = allocation.height - 1

        if self.top:
            context.move_to(x_loc, y_loc)
            context.line_to(x_loc + width, y_loc)
        if self.right:
            context.move_to(x_loc + width, y_loc)
            context.line_to(x_loc + width, y_loc + height)
        if self.bottom:
            context.move_to(x_loc + width, y_loc + height)
            context.line_to(x_loc, y_loc + height)
        if self.left:
            context.move_to(x_loc, y_loc + height)
            context.line_to(x_loc, y_loc)
        context.set_line_width(1)
        context.stroke()

    def _updateBorders(self):
        self.set_padding(
            self.top and 1 or 0,
            self.bottom and 1 or 0,
            self.right and 1 or 0,
            self.left and 1 or 0,
        )

    def isTop(self):
        return self.__top

    def isRight(self):
        return self.__right

    def isBottom(self):
        return self.__bottom

    def isLeft(self):
        return self.__left

    def setTop(self, value):
        self.__top = value
        self._updateBorders()

    def setRight(self, value):
        self.__right = value
        self._updateBorders()

    def setBottom(self, value):
        self.__bottom = value
        self._updateBorders()

    def setLeft(self, value):
        self.__left = value
        self._updateBorders()

    top = property(isTop, setTop, None, None)

    right = property(isRight, setRight, None, None)

    bottom = property(isBottom, setBottom, None, None)

    left = property(isLeft, setLeft, None, None)



from pychess.System import conf
from pychess.ic.managers.ListAndVarManager import ListAndVarManager


class ICCListAndVarManager(ListAndVarManager):
    def __init__(self, connection):
        self.connection = connection

        self.publicLists = {}
        self.personalLists = {}
        self.personalBackup = {}

        # Auto flag
        conf.notify_add("autoCallFlag", self.autoFlagNotify)

    def autoFlagNotify(self, *args):
        self.connection.client.run_command(
            "set autoflag %s" % int(conf.get("autoCallFlag"))
        )
        # print 'notify flag', conf.get('autoCallFlag')



"""The Racing Kings Variation"""

from pychess.Utils.const import (
    RACINGKINGSCHESS,
    VARIANTS_OTHER_NONSTANDARD,
    A8,
    B8,
    C8,
    D8,
    E8,
    F8,
    G8,
    H8,
)
from pychess.Utils.Board import Board

RACINGKINGSSTART = "8/8/8/8/8/8/krbnNBRK/qrbnNBRQ w - - 0 1"

RANK8 = (A8, B8, C8, D8, E8, F8, G8, H8)


class RacingKingsBoard(Board):
    """:Description: The Racing Kings variation is where the object of the game
    is to bring your king to the eight row.
    """

    variant = RACINGKINGSCHESS
    __desc__ = _(
        "In this game, check is entirely forbidden: not only is it forbidden\n"
        + "to move ones king into check, but it is also forbidden to check the opponents king.\n"
        + "The purpose of the game is to be the first player that moves his king to the eight row.\n"
        + "When white moves their king to the eight row, and black moves directly after that also\n"
        + "their king to the last row, the game is a draw\n"
        + "(this rule is to compensate for the advantage of white that they may move first.)\n"
        + "Apart from the above, pieces move and capture precisely as in normal chess."
    )
    name = _("Racing Kings")
    cecp_name = "racingkings"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=RACINGKINGSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


def testKingInEightRow(board):
    """Test for a winning position"""
    return board.kings[board.color - 1] in RANK8


def test2KingInEightRow(board):
    """Test for a winning position"""
    return board.kings[board.color] in RANK8 and board.kings[board.color - 1] in RANK8



import asyncio
import os
import sys
import pstats
import inspect
import cProfile
import ssl
import tempfile
from timeit import default_timer
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from pychess import MSYS2

# TODO: this hack only needed until we can figure out
# what ssl cert files needed to copied in setup.py
# when cx_freezee bdist_msi command is used
if MSYS2:
    ssl._create_default_https_context = ssl._create_unverified_context


async def cancel_all_tasks():
    loop = asyncio.get_event_loop()
    tasks = [
        task for task in asyncio.all_tasks(loop) if task is not asyncio.current_task()
    ]
    for task in tasks:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            print("task cancelled", task)


async def download_file_async(url, progressbar=None):
    loop = asyncio.get_event_loop()
    future = loop.run_in_executor(None, download_file, url)
    temp_file = await future
    return temp_file


def download_file(url, progressbar=None):
    temp_file = None
    try:
        if progressbar is not None:
            from gi.repository import GLib

            GLib.idle_add(progressbar.set_text, "Downloading %s ..." % url)
        else:
            print("Downloading %s ..." % url)

        header = {"User-Agent": "Mozilla/5.0 (compatible)"}
        req = Request(url, headers=header)
        f = urlopen(req)

        temp_file = os.path.join(tempfile.gettempdir(), os.path.basename(url))
        with open(temp_file, "wb") as local_file:
            local_file.write(f.read())

    except HTTPError as e:
        print("HTTP Error:", e.code, url)

    except URLError as e:
        print("URL Error:", e.reason, url)

    return temp_file


def fident(f):
    """
    Get an identifier for a function or method
    """
    joinchar = "."
    if hasattr(f, "im_class"):
        fparent = f.im_class.__name__
    else:
        joinchar = ":"
        fparent = f.__module__.split(".")[-1]

    # sometimes inspect.getsourcelines() segfaults on windows
    if getattr(sys, "frozen", False) or sys.platform == "win32":
        lineno = 0
    else:
        lineno = inspect.getsourcelines(f)[1]

    fullname = joinchar.join((fparent, f.__name__))
    return ":".join((fullname, str(lineno)))


def get_threadname(thread_namer):
    if isinstance(thread_namer, str):
        return thread_namer
    else:
        return fident(thread_namer)


# https://gist.github.com/techtonik/2151727
def caller_name(skip=2):
    """Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed stack height
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ""
    parentframe = stack[start][0]

    name = []
    module = inspect.getmodule(parentframe)
    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)
    # detect classname
    if "self" in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parentframe.f_locals["self"].__class__.__name__)
    codename = parentframe.f_code.co_name
    if codename != "<module>":  # top level usually
        name.append(codename)  # function or a method
    del parentframe
    return ".".join(name)


def profile_me(fn):
    def profiled_fn(*args, **kwargs):
        prof = cProfile.Profile()
        ret = prof.runcall(fn, *args, **kwargs)
        ps = pstats.Stats(prof)
        ps.sort_stats("cumulative")
        ps.print_stats(60)
        return ret

    return profiled_fn


# Python Timer Class - Context Manager for Timing Code Blocks
# Corey Goldberg - 2012
class Timer:
    def __init__(self, text):
        self.text = text
        self.timer = default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        end = self.timer()
        self.elapsed_secs = end - self.start
        self.elapsed = self.elapsed_secs * 1000  # millisecs
        print(f"---- elapsed time: {self.elapsed:f} ms - {self.text}")



import sys
import email.utils
import math
import random
import signal
import subprocess
from threading import Thread
from urllib.parse import urlencode
from urllib.request import urlopen

from gi.repository import Gtk, Gdk

import pychess
from pychess.Players.PyChess import PyChess
from pychess.System.repeat import repeat_sleep
from pychess.System import fident
from pychess.System.Log import log
from pychess.Utils.const import (
    WHITE,
    PAUSE_OFFER,
    DRAW_OFFER,
    BLACK,
    RESUME_OFFER,
    TAKEBACK_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    SWITCH_OFFER,
    NORMALCHESS,
    SAN,
)
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.lmove import determineAlgebraicNotation, toLAN, parseSAN
from pychess.Utils.lutils import lsearch
from pychess.Utils.repr import reprResult_long, reprReason_long
from pychess.ic.FICSConnection import FICSMainConnection


class PyChessFICS(PyChess):
    def __init__(self, password, from_address, to_address):
        PyChess.__init__(self)

        self.ports = (23, 5000)
        if not password:
            self.username = "guest"
        else:
            self.username = "PyChess"
        self.owner = "Lobais"
        self.password = password
        self.from_address = "The PyChess Bot <%s>" % from_address
        self.to_address = "Thomas Dybdahl Ahle <%s>" % to_address

        # Possible start times
        self.minutes = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
        self.gains = (0, 5, 10, 15, 20)
        # Possible colors. None == random
        self.colors = (WHITE, BLACK, None)
        # The amount of random challenges, that PyChess sends with each seek
        self.challenges = 10
        # enableEGTB()

        self.sudos = set()
        self.ownerOnline = False
        self.waitingForPassword = None
        self.log = []
        self.acceptedTimesettings = []

        self.worker = None

        repeat_sleep(self.sendChallenges, 60 * 1)

    def __triangular(self, low, high, mode):
        """Triangular distribution.
        Continuous distribution bounded by given lower and upper limits,
        and having a given mode value in-between.
        http://en.wikipedia.org/wiki/Triangular_distribution
        """
        random_num = random.random()
        mid = (mode - low) / (high - low)
        if random_num > mid:
            random_num = 1 - random_num
            mid = 1 - mid
            low, high = high, low
        tri = low + (high - low) * (random_num * mid) ** 0.5
        if tri < mode:
            return int(tri)
        elif tri > mode:
            return int(math.ceil(tri))
        return int(round(tri))

    def sendChallenges(self):
        if self.connection.bm.isPlaying():
            return True

        statsbased = (
            (0.39197722779282, 3, 0),
            (0.59341408108783, 5, 0),
            (0.77320877377846, 1, 0),
            (0.8246379941394, 10, 0),
            (0.87388717406441, 2, 12),
            (0.91443760169489, 15, 0),
            (0.9286423058163, 4, 0),
            (0.93891977227793, 2, 0),
            (0.94674539138335, 20, 0),
            (0.95321476842423, 2, 2),
            (0.9594588808257, 5, 2),
            (0.96564528079889, 3, 2),
            (0.97173859621034, 7, 0),
            (0.97774906636184, 3, 1),
            (0.98357243654425, 5, 12),
            (0.98881309737017, 5, 5),
            (0.99319644938247, 6, 0),
            (0.99675879556023, 3, 12),
            (1, 5, 3),
        )

        # n = random.random()
        # for culminativeChance, minute, gain in statsbased:
        #    if n < culminativeChance:
        #        break

        culminativeChance, minute, gain = random.choice(statsbased)

        # type = random.choice((TYPE_LIGHTNING, TYPE_BLITZ, TYPE_STANDARD))
        # if type == TYPE_LIGHTNING:
        #    minute = self.__triangular(0,2+1,1)
        #    mingain = not minute and 1 or 0
        #    maxgain = int((3-minute)*3/2)
        #    gain = random.randint(mingain, maxgain)
        # elif type == TYPE_BLITZ:
        #    minute = self.__triangular(0,14+1,5)
        #    mingain = max(int((3-minute)*3/2+1), 0)
        #    maxgain = int((15-minute)*3/2)
        #    gain = random.randint(mingain, maxgain)
        # elif type == TYPE_STANDARD:
        #    minute = self.__triangular(0,20+1,12)
        #    mingain = max(int((15-minute)*3/2+1), 0)
        #    maxgain = int((20-minute)*3/2)
        #    gain = self.__triangular(mingain, maxgain, mingain)

        # color = random.choice(self.colors)
        self.extendlog(["Seeking %d %d" % (minute, gain)])
        self.connection.glm.seek(minute, gain, True)
        opps = random.sample(
            self.connection.players.get_online_playernames(), self.challenges
        )
        self.extendlog("Challenging %s" % op for op in opps)
        for player in opps:
            self.connection.om.challenge(player, minute, gain, True)

        return True

    def makeReady(self):
        signal.signal(signal.SIGINT, Gtk.main_quit)

        PyChess.makeReady(self)

        self.connection = FICSMainConnection(
            "freechess.org", self.ports, True, self.username, self.password
        )
        self.connection.connect("connectingMsg", self.__showConnectLog)
        self.connection._connect()

        self.connection.glm.connect("addPlayer", self.__onAddPlayer)
        self.connection.glm.connect("removePlayer", self.__onRemovePlayer)
        self.connection.cm.connect("privateMessage", self.__onTell)
        self.connection.alm.connect("logOut", self.__onLogOut)
        self.connection.bm.connect("playGameCreated", self.__onGameCreated)
        self.connection.bm.connect("curGameEnded", self.__onGameEnded)
        self.connection.bm.connect("boardUpdate", self.__onBoardUpdate)
        self.connection.om.connect("onChallengeAdd", self.__onChallengeAdd)
        self.connection.om.connect("onOfferAdd", self.__onOfferAdd)
        self.connection.adm.connect("onAdjournmentsList", self.__onAdjournmentsList)
        self.connection.em.connect("onAmbiguousMove", self.__onAmbiguousMove)
        self.connection.em.connect("onIllegalMove", self.__onAmbiguousMove)

        self.connection.adm.queryAdjournments()
        self.connection.lvm.setVariable("autoflag", 1)

        self.connection.fm.setFingerNote(
            1,
            "PyChess is the chess engine bundled with the PyChess %s " % pychess.VERSION
            + "chess client. This instance is owned by %s, but acts " % self.owner
            + "quite autonomously.",
        )

        self.connection.fm.setFingerNote(
            2,
            "PyChess is 100% Python code and is released under the terms of "
            + "the GPL. The evalution function is largely equal to the one of"
            + "GnuChess, but it plays quite differently.",
        )

        self.connection.fm.setFingerNote(
            3,
            "PyChess runs on an elderly AMD Sempron(tm) Processor 3200+, 512 "
            + "MB DDR2 Ram, but is built to take use of 64bit calculating when "
            + "accessible, through the gpm library.",
        )

        self.connection.fm.setFingerNote(
            4,
            "PyChess uses a small 500 KB openingbook based solely on Kasparov "
            + "games. The engine doesn't have much endgame knowledge, but might "
            + "in some cases access an online endgamedatabase.",
        )

        self.connection.fm.setFingerNote(
            5,
            "PyChess will allow any pause/resume and adjourn wishes, but will "
            + "deny takebacks. Draw, abort and switch offers are accepted, "
            + "if they are found to be an advance. Flag is auto called, but "
            + "PyChess never resigns. We don't want you to forget your basic "
            + "mating skills.",
        )

    def main(self):
        self.connection.run()
        self.extendlog([str(self.acceptedTimesettings)])
        self.phoneHome("Session ended\n" + "\n".join(self.log))
        print("Session ended")

    def run(self):
        t = Thread(target=self.main, name=fident(self.main))
        t.daemon = True
        t.start()
        Gdk.threads_init()
        Gtk.main()

    # General

    def __showConnectLog(self, connection, message):
        print(message)

    def __onLogOut(self, autoLogoutManager):
        self.connection.close()
        # sys.exit()

    def __onAddPlayer(self, gameListManager, player):
        if player["name"] in self.sudos:
            self.sudos.remove(player["name"])
        if player["name"] == self.owner:
            self.connection.cm.tellPlayer(self.owner, "Greetings")
            self.ownerOnline = True

    def __onRemovePlayer(self, gameListManager, playername):
        if playername == self.owner:
            self.ownerOnline = False

    def __onAdjournmentsList(self, adjournManager, adjournments):
        for adjournment in adjournments:
            if adjournment["online"]:
                adjournManager.challenge(adjournment["opponent"])

    def __usage(self):
        return (
            "|| PyChess bot help file || "
            + "# help 'Displays this help file' "
            + "# sudo <password> <command> 'Lets PyChess execute the given command' "
            + "# sendlog 'Makes PyChess send you its current log'"
        )

    def __onTell(self, chatManager, name, title, isadmin, text):
        if self.waitingForPassword:
            if text.strip() == self.password or (not self.password and text == "none"):
                self.sudos.add(name)
                self.tellHome("%s gained sudo access" % name)
                self.connection.client.run_command(self.waitingForPassword)
            else:
                chatManager.tellPlayer(name, "Wrong password")
                self.tellHome("%s failed sudo access" % name)
            self.waitingForPassword = None
            return

        args = text.split()

        # if args == ["help"]:
        #    chatManager.tellPlayer(name, self.__usage())

        if args[0] == "sudo":
            command = " ".join(args[1:])
            if name in self.sudos or name == self.owner:
                # Notice: This can be used to make nasty loops
                print(command, file=self.connection.client)
            else:
                print(repr(name), self.sudos)
                chatManager.tellPlayer(name, "Please send me the password")
                self.waitingForPassword = command

        elif args == ["sendlog"]:
            if self.log:
                # TODO: Consider email
                chatManager.tellPlayer(name, "\\n".join(self.log))
            else:
                chatManager.tellPlayer(name, "The log is currently empty")

        else:
            if self.ownerOnline:
                self.tellHome(f"{name} told me '{text}'")
            else:

                def onlineanswer(message):
                    data = (
                        urlopen(
                            "http://www.pandorabots.com/pandora/talk?botid=8d034368fe360895",
                            urlencode({"message": message, "botcust2": "x"}).encode(
                                "utf-8"
                            ),
                        )
                        .read()
                        .decode("utf-8")
                    )
                    bold_ss = "<b>DMPGirl:</b>"
                    break_es = "<br>"
                    answer = data[
                        data.find(bold_ss)
                        + len(bold_ss) : data.find(break_es, data.find(bold_ss))
                    ]
                    chatManager.tellPlayer(name, answer)

                thread = Thread(
                    target=onlineanswer, name=fident(onlineanswer), args=(text,)
                )
                thread.daemon = True
                thread.start()
            # chatManager.tellPlayer(name, "Sorry, your request was nonsense.\n"+\
            # "Please read my help file for more info")

            # Challenges and other offers

    def __onChallengeAdd(self, offerManager, index, match):
        # match = {"tp": type, "w": fname, "rt": rating, "color": color,
        #         "r": rated, "t": mins, "i": incr}
        offerManager.acceptIndex(index)

    def __onOfferAdd(self, offerManager, offer):
        if offer.type in (PAUSE_OFFER, RESUME_OFFER, ADJOURN_OFFER):
            offerManager.accept(offer)
        elif offer.type in (TAKEBACK_OFFER,):
            offerManager.decline(offer)
        elif offer.type in (DRAW_OFFER, ABORT_OFFER, SWITCH_OFFER):
            if self.__willingToDraw():
                offerManager.accept(offer)
            else:
                offerManager.decline(offer)

    # Playing

    def __onGameCreated(self, boardManager, ficsgame):
        base = int(ficsgame.minutes) * 60
        inc = int(ficsgame.inc)
        self.clock[:] = base, base
        self.increment[:] = inc, inc
        self.gameno = ficsgame.gameno
        self.lastPly = -1

        self.acceptedTimesettings.append((base, inc))

        self.tellHome(
            "Starting a game (%s, %s) gameno: %s"
            % (ficsgame.wplayer.name, ficsgame.bplayer.name, ficsgame.gameno)
        )

        if ficsgame.bplayer.name.lower() == self.connection.getUsername().lower():
            self.playingAs = BLACK
        else:
            self.playingAs = WHITE

        self.board = LBoard(NORMALCHESS)
        # Now we wait until we receive the board.

    def __go(self):
        if self.worker:
            self.worker.cancel()
        # TODO: fix it
        # self.worker = GtkWorker(
        #     lambda worker: PyChess._PyChess__go(self, worker))
        # self.worker.connect("published", lambda w, msg: self.extendlog(msg))
        # self.worker.connect("done", self.__onMoveCalculated)
        # self.worker.execute()

    def __willingToDraw(self):
        return self.scr <= 0  # FIXME: this misbehaves in all but the simplest use cases

    def __onGameEnded(self, boardManager, ficsgame):
        self.tellHome(
            reprResult_long[ficsgame.result] + " " + reprReason_long[ficsgame.reason]
        )
        lsearch.searching = False
        if self.worker:
            self.worker.cancel()
            self.worker = None

    def __onMoveCalculated(self, worker, sanmove):
        if worker.isCancelled() or not sanmove:
            return
        self.board.applyMove(parseSAN(self.board, sanmove))
        self.connection.bm.sendMove(sanmove)
        self.extendlog(["Move sent %s" % sanmove])

    def __onBoardUpdate(
        self, boardManager, gameno, ply, curcol, lastmove, fen, wname, bname, wms, bms
    ):
        self.extendlog(["", "I got move %d %s for gameno %s" % (ply, lastmove, gameno)])

        if self.gameno != gameno:
            return

        self.board.applyFen(fen)

        self.clock[:] = wms / 1000.0, bms / 1000.0

        if curcol == self.playingAs:
            self.__go()

    def __onAmbiguousMove(self, errorManager, move):
        # This is really a fix for fics, but sometimes it is necessary
        if determineAlgebraicNotation(move) == SAN:
            self.board.popMove()
            move_ = parseSAN(self.board, move)
            lanmove = toLAN(self.board, move_)
            self.board.applyMove(move_)
            self.connection.bm.sendMove(lanmove)
        else:
            self.connection.cm.tellOpponent(
                "I'm sorry, I wanted to move %s, but FICS called " % move
                + "it 'Ambigious'. I can't find another way to express it, "
                + "so you can win"
            )
            self.connection.bm.resign()

    # Utils

    def extendlog(self, messages):
        [log.info(m + "\n") for m in messages]
        self.log.extend(messages)
        del self.log[:-10]

    def tellHome(self, message):
        print(message)
        if self.ownerOnline:
            self.connection.cm.tellPlayer(self.owner, message)

    def phoneHome(self, message):
        SENDMAIL = "/usr/sbin/sendmail"
        SUBJECT = "Besked fra botten"

        pipe = subprocess.Popen(
            [
                SENDMAIL,
                "-f",
                email.utils.parseaddr(self.from_address)[1],
                email.utils.parseaddr(self.to_address)[1],
            ],
            stdin=subprocess.PIPE,
        )

        print("MIME-Version: 1.0", file=pipe.stdin)
        print("Content-Type: text/plain; charset=UTF-8", file=pipe.stdin)
        print("Content-Disposition: inline", file=pipe.stdin)
        print("From: %s" % self.from_address, file=pipe.stdin)
        print("To: %s" % self.to_address, file=pipe.stdin)
        print("Subject: %s" % SUBJECT, file=pipe.stdin)
        print(file=pipe.stdin)
        print(message, file=pipe.stdin)
        print("Cheers", file=pipe.stdin)

        pipe.stdin.close()
        pipe.wait()


if __name__ == "__main__":
    if len(sys.argv) == 5 and sys.argv[1] == "fics":
        pychess_fics = PyChessFICS(*sys.argv[2:])
    else:
        print("Unknown argument(s):", repr(sys.argv))
        sys.exit(0)

    pychess_fics.makeReady()
    pychess_fics.run()



import os
import random
import sys
from time import time

this_dir = os.path.dirname(os.path.abspath(__file__))
if os.path.join(this_dir, "../..") not in sys.path:
    sys.path = [os.path.join(this_dir, "../..")] + sys.path

try:
    from pychess.Utils.book import getOpenings  # nopep8
    from pychess.Utils.const import (
        WHITE,
        ASEANCHESS,
        SITTUYINCHESS,
        ATOMICCHESS,
        reprResult,
        CAMBODIANCHESS,
        LOSERSCHESS,
        KINGOFTHEHILLCHESS,
        DRAW,
        BLACKWON,
        WHITEWON,
        MAKRUKCHESS,
        SUICIDECHESS,
        GIVEAWAYCHESS,
        THREECHECKCHESS,
        HORDECHESS,
        RACINGKINGSCHESS,
        PLACEMENTCHESS,
    )  # nopep8
    from pychess.Utils.lutils import lsearch  # nopep8
    from pychess.Utils.lutils.ldata import MAXPLY  # nopep8
    from pychess.Utils.lutils.lsearch import alphaBeta  # nopep8
    from pychess.Utils.lutils.lmove import listToSan, toSAN  # nopep8
    from pychess.System.Log import log  # nopep8
except ImportError:
    print("ERROR: failed to start PyChess.py engine")
    sys.exit(1)


class PyChess:
    def __init__(self):
        self.sd = MAXPLY
        self.skipPruneChance = 0

        self.clock = [0, 0]
        self.basetime = 0
        self.increment = 0
        self.movestogo = 0
        self.searchtime = 0
        self.scr = 0  # The current predicted score. Used when accepting draw offers
        self.playingAs = WHITE
        self.ponder = False  # Currently unused
        self.post = False
        self.debug = True
        self.outOfBook = False

    def print(self, text):
        try:
            print(text, flush=True)
            log.debug(text, extra={"task": "stdout"})
        except BrokenPipeError:
            sys.exit(0)

    # Play related

    def __remainingMovesA(self):
        # Based on regression of a 180k games pgn
        ply_count = self.board.plyCount
        remaining = (
            -1.71086e-12 * ply_count**6
            + 1.69103e-9 * ply_count**5
            - 6.00801e-7 * ply_count**4
            + 8.17741e-5 * ply_count**3
            + 2.91858e-4 * ply_count**2
            - 0.94497 * ply_count
            + 78.8979
        )
        self.print("# remaining moves estimate=%s" % remaining)
        return remaining

    def __remainingMovesB(self):
        # Classical timecontrol
        ply = self.board.plyCount % (self.movestogo * 2)
        remaining = self.movestogo - ply // 2
        self.print("# remaining moves estimate=%s" % remaining)
        return remaining

    def __getBestOpening(self):
        totalWeight = 0
        choice = None
        if self.board.variant not in (
            ASEANCHESS,
            CAMBODIANCHESS,
            MAKRUKCHESS,
            HORDECHESS,
            PLACEMENTCHESS,
            SITTUYINCHESS,
            LOSERSCHESS,
            SUICIDECHESS,
            GIVEAWAYCHESS,
            ATOMICCHESS,
            KINGOFTHEHILLCHESS,
            THREECHECKCHESS,
            RACINGKINGSCHESS,
        ):
            for move, weight, learn in getOpenings(self.board):
                totalWeight += weight
                if totalWeight == 0:
                    break
                if not move or random.randrange(totalWeight) < weight:
                    choice = move
        if choice is None:
            self.outOfBook = True
        return choice

    def __go(self, ondone=None):
        """Finds and prints the best move from the current position"""

        mv = False if self.outOfBook else self.__getBestOpening()
        if mv:
            mvs = [mv]

        if not mv:
            lsearch.skipPruneChance = self.skipPruneChance
            lsearch.searching = True

            timed = self.basetime > 0 or self.increment > 0 or self.searchtime > 0

            if self.searchtime > 0:
                usetime = self.searchtime
            else:
                if self.movestogo > 0:
                    remaining_moves = self.__remainingMovesB()
                    usetime = self.clock[self.playingAs] / remaining_moves
                    if remaining_moves == 1:
                        usetime -= 0.05
                else:
                    usetime = self.clock[self.playingAs] / self.__remainingMovesA()
                    if self.clock[self.playingAs] > 10:
                        # If we have time, we assume 40 moves rather than 80
                        usetime *= 2
                    # The increment is a constant. We'll use this always
                    usetime += self.increment
                usetime = min(usetime, self.clock[self.playingAs] / 2)

            prevtime = 0
            starttime = time()
            lsearch.endtime = starttime + usetime if timed else sys.maxsize
            if self.debug:
                if timed:
                    self.print(
                        "# Time left: %3.2f s; Planing to think for %3.2f s"
                        % (self.clock[self.playingAs], usetime)
                    )
                else:
                    self.print("# Searching to depth %d without timelimit" % self.sd)

            for depth in range(1, self.sd + 1):
                # Heuristic time saving
                # Don't waste time, if the estimated isn't enough to complete
                # next depth
                if depth > 1 and timed and usetime <= prevtime * 4 and usetime > 1:
                    break
                lsearch.timecheck_counter = lsearch.TIMECHECK_FREQ
                search_result = alphaBeta(self.board, depth)
                if lsearch.searching:
                    mvs, self.scr = search_result
                    if time() > lsearch.endtime:
                        break
                    if self.post:
                        pv1 = " ".join(listToSan(self.board, mvs))
                        time_cs = int(100 * (time() - starttime))
                        self.print(
                            "{} {} {} {} {}".format(
                                depth, self.scr, time_cs, lsearch.nodes, pv1
                            )
                        )
                else:
                    # We were interrupted
                    if depth == 1:
                        mvs, self.scr = search_result
                    break
                prevtime = time() - starttime - prevtime

                self.clock[self.playingAs] -= time() - starttime - self.increment

            if not mvs:
                if not lsearch.searching:
                    # We were interupted
                    lsearch.nodes = 0
                    return

                # This should only happen in terminal mode

                if self.scr == 0:
                    self.print("result %s" % reprResult[DRAW])
                elif self.scr < 0:
                    if self.board.color == WHITE:
                        self.print("result %s" % reprResult[BLACKWON])
                    else:
                        self.print("result %s" % reprResult[WHITEWON])
                else:
                    if self.board.color == WHITE:
                        self.print("result %s" % reprResult[WHITEWON])
                    else:
                        self.print("result %s" % reprResult[BLACKWON])
                return

            lsearch.nodes = 0
            lsearch.searching = False

        move = mvs[0]
        sanmove = toSAN(self.board, move)
        if ondone:
            ondone(sanmove)
        return sanmove

    def __analyze(self):
        """Searches, and prints info from, the position as stated in the cecp
        protocol"""

        start = time()
        lsearch.endtime = sys.maxsize
        lsearch.searching = True

        for depth in range(1, self.sd):
            if not lsearch.searching:
                break
            board = self.board.clone()
            mvs, scr = alphaBeta(board, depth)

            pv1 = " ".join(listToSan(board, mvs))
            time_cs = int(100 * (time() - start))
            self.print(f"{depth} {scr} {time_cs} {lsearch.nodes} {pv1}")

            lsearch.nodes = 0


if __name__ == "__main__":
    import logging
    from pychess.Players.PyChessCECP import PyChessCECP

    if len(sys.argv) == 1 or sys.argv[1:] == ["debug"]:
        if "debug" in sys.argv[1:]:
            log.logger.setLevel(logging.DEBUG)
        else:
            log.logger.setLevel(logging.WARNING)

        pychess = PyChessCECP()
    else:
        print("Unknown argument(s):", repr(sys.argv))
        sys.exit(0)

    pychess.makeReady()
    pychess.run()



from gi.repository import GLib

from pychess.System import conf
from pychess.Utils.const import ACTION_MENU_ITEMS

################################################################################
# Main menubar MenuItem classes to keep track of menu widget states            #
################################################################################


class GtkMenuItem:
    def __init__(self, name, sensitive=False, label=None, tooltip=None):
        assert isinstance(sensitive, bool)
        assert label is None or isinstance(label, str)
        self.name = name
        self._sensitive = sensitive
        self._label = label
        self._tooltip = tooltip

    @property
    def sensitive(self):
        return self._sensitive

    @sensitive.setter
    def sensitive(self, sensitive):
        assert isinstance(sensitive, bool)
        self._sensitive = sensitive
        self._set_widget("sensitive", sensitive)

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        assert isinstance(label, str)
        self._label = label
        self._set_widget("label", label)

    @property
    def tooltip(self):
        return self._tooltip

    @tooltip.setter
    def tooltip(self, tooltip):
        assert isinstance(tooltip, str)
        self._tooltip = tooltip
        self._set_widget("tooltip-text", tooltip)

    def _set_widget(self, prop, value):
        from . import gamewidget

        if gamewidget.getWidgets()[self.name].get_property(prop) != value:
            # print("setting %s property %s to %s.." % (self.name, prop, str(value)))

            def do_set_menu_item_prop():
                gamewidget.getWidgets()[self.name].set_property(prop, value)

            GLib.idle_add(do_set_menu_item_prop)
            # print(" success (%s %s = \"%s\")" % (
            #     self.name, prop, gamewidget.getWidgets()[self.name].get_property(prop)))

    def update(self):
        self._set_widget("sensitive", self._sensitive)
        if self._label is not None:
            self._set_widget("label", self._label)
        if self._tooltip is not None:
            self._set_widget("tooltip-text", self._tooltip)


class GtkMenuToggleButton(GtkMenuItem):
    def __init__(self, name, sensitive=False, active=False, label=None):
        assert isinstance(active, bool)
        GtkMenuItem.__init__(self, name, sensitive, label)
        self._active = active

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, active):
        assert isinstance(active, bool)
        self._active = active
        self._set_widget("active", active)

    def update(self):
        GtkMenuItem.update(self)
        self._set_widget("active", self._active)


class MenuItemsDict(dict):
    """
    :Description: Keeps track of menubar menuitem widgets that need to be managed on a game
    by game basis. Each menuitem writes through its respective widget state to
    the GUI if we are encapsulated in the gamewidget that's focused/infront
    """

    ANAL_MENU_ITEMS = (
        "analyze_game1",
        "analyzer_check",
        "inv_analyzer_check",
        "ana_combobox",
        "inv_ana_combobox",
    )
    VIEW_MENU_ITEMS = ("hint_mode", "spy_mode")

    def __init__(self):
        dict.__init__(self)
        for item in ACTION_MENU_ITEMS:
            dict.__setitem__(self, item, GtkMenuItem(item))
        for item in self.ANAL_MENU_ITEMS:
            dict.__setitem__(self, item, GtkMenuItem(item, sensitive=True))
        for item in self.VIEW_MENU_ITEMS:
            dict.__setitem__(
                self, item, GtkMenuToggleButton(item, active=conf.get(item))
            )



import asyncio
from io import StringIO

from gi.repository import Gtk, GObject

from pychess.Utils.const import (
    DRAW,
    LOCAL,
    WHITE,
    BLACK,
    WAITING_TO_START,
    reprResult,
    UNDOABLE_STATES,
    FIRST_PAGE,
    PREV_PAGE,
    NEXT_PAGE,
)
from pychess.Players.Human import Human
from pychess.Utils.GameModel import GameModel
from pychess.perspectives import perspective_manager
from pychess.Variants import variants
from pychess.Database.model import game, event, site, pl1, pl2
from pychess.widgets import newGameDialog
from pychess.Savers import pgn


cols = (
    game.c.id,
    pl1.c.name,
    game.c.white_elo,
    pl2.c.name,
    game.c.black_elo,
    game.c.result,
    game.c.date,
    event.c.name,
    site.c.name,
    game.c.round,
    game.c.ply_count,
    game.c.eco,
    game.c.time_control,
    game.c.variant,
    game.c.fen,
)


class GameList(Gtk.TreeView):
    def __init__(self, persp):
        GObject.GObject.__init__(self)
        self.persp = persp

        self.records = []
        self.preview_cid = None

        # GTK_SELECTION_BROWSE - exactly one item is always selected
        self.get_selection().set_mode(Gtk.SelectionMode.BROWSE)

        self.liststore = Gtk.ListStore(
            int, str, str, str, str, str, str, str, str, str, str, str, str, str, str
        )
        self.modelsort = Gtk.TreeModelSort(self.liststore)

        self.modelsort.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.modelsort.set_sort_func(2, self.compare_elo, 2)
        self.modelsort.set_sort_func(4, self.compare_elo, 4)
        self.modelsort.connect("sort-column-changed", self.sort_column_changed)
        self.set_model(self.modelsort)
        self.get_selection().set_mode(Gtk.SelectionMode.BROWSE)
        self.set_headers_visible(True)
        self.set_rules_hint(True)
        self.set_search_column(1)

        titles = (
            _("Id"),
            _("White"),
            _("W Elo"),
            _("Black"),
            _("B Elo"),
            _("Result"),
            _("Date"),
            _("Event"),
            _("Site"),
            _("Round"),
            _("Length"),
            "ECO",
            _("Time control"),
            _("Variant"),
            "FEN",
        )

        for i, title in enumerate(titles):
            r = Gtk.CellRendererText()
            if i in (0, 2, 4, 9, 10, 12):
                r.set_alignment(
                    1.0, 0.5
                )  # set alignment to the right for numeric values
            column = Gtk.TreeViewColumn(title, r, text=i)
            column.set_resizable(True)
            column.set_reorderable(True)
            column.set_sort_column_id(i)
            self.append_column(column)

        self.connect("row-activated", self.row_activated)

        self.set_cursor(0)
        self.columns_autosize()
        self.gamemodel = GameModel()
        self.ply = 0

        # buttons
        toolbar = Gtk.Toolbar()

        firstButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_PREVIOUS)
        firstButton.set_tooltip_text(_("First games"))
        toolbar.insert(firstButton, -1)

        prevButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_REWIND)
        prevButton.set_tooltip_text(_("Previous games"))
        toolbar.insert(prevButton, -1)

        nextButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_FORWARD)
        nextButton.set_tooltip_text(_("Next games"))
        toolbar.insert(nextButton, -1)

        firstButton.connect("clicked", self.on_first_clicked)
        prevButton.connect("clicked", self.on_prev_clicked)
        nextButton.connect("clicked", self.on_next_clicked)

        limit_combo = Gtk.ComboBoxText()
        for limit in ("100", "500", "1000", "5000"):
            limit_combo.append_text(limit)
        limit_combo.set_active(0)

        toolitem = Gtk.ToolItem.new()
        toolitem.add(limit_combo)
        toolbar.insert(toolitem, -1)
        limit_combo.connect("changed", self.on_limit_combo_changed)

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        sw.add(self)

        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.box.pack_start(sw, True, True, 0)
        self.box.pack_start(toolbar, False, False, 0)
        self.box.show_all()

    def on_first_clicked(self, button):
        self.load_games(direction=FIRST_PAGE)

    def on_prev_clicked(self, button):
        self.load_games(direction=PREV_PAGE)

    def on_next_clicked(self, button):
        self.load_games(direction=NEXT_PAGE)

    def on_limit_combo_changed(self, combo):
        text = combo.get_active_text()
        if text is not None:
            self.persp.chessfile.limit = int(text)
            self.load_games(direction=FIRST_PAGE)

    def compare_elo(self, treesortable, iter0, iter1, column):
        """Adapted sort : int values are not sorted in lexicographic order"""
        elo0 = treesortable.get_value(iter0, column)
        elo1 = treesortable.get_value(iter1, column)

        isnum0 = elo0.isdigit()
        isnum1 = elo1.isdigit()

        if isnum0 and isnum1:
            num0 = int(elo0)
            num1 = int(elo1)
            return (num0 > num1) - (num0 < num1)
        elif isnum0:
            return 1
        elif isnum1:
            return -1
        else:
            return (elo0 > elo1) - (elo0 < elo1)

    def sort_column_changed(self, treesortable):
        sort_column_id, order = treesortable.get_sort_column_id()
        if sort_column_id is None:
            self.modelsort.set_sort_column_id(0, Gtk.SortType.ASCENDING)
            sort_column_id, order = 0, Gtk.SortType.ASCENDING

        self.set_search_column(sort_column_id)
        is_desc = order == Gtk.SortType.DESCENDING
        self.persp.chessfile.set_tag_order(cols[sort_column_id], is_desc)
        self.load_games(direction=FIRST_PAGE)

    def load_games(self, direction=FIRST_PAGE):
        selection = self.get_selection()
        if (
            selection is not None
            and self.preview_cid is not None
            and selection.handler_is_connected(self.preview_cid)
        ):
            with GObject.signal_handler_block(selection, self.preview_cid):
                self.liststore.clear()
        else:
            self.liststore.clear()

        add = self.liststore.append

        self.records = []
        records, plys = self.persp.chessfile.get_records(direction)
        for i, rec in enumerate(records):
            game_id = rec["Id"]
            offs = rec["Offset"]
            wname = rec["White"]
            bname = rec["Black"]
            welo = rec["WhiteElo"]
            belo = rec["BlackElo"]
            result = rec["Result"]
            result = "½-½" if result == DRAW else reprResult[result] if result else "*"
            event = "" if rec["Event"] is None else rec["Event"].replace("?", "")
            site = "" if rec["Site"] is None else rec["Site"].replace("?", "")
            round_ = "" if rec["Round"] is None else rec["Round"].replace("?", "")
            date = (
                ""
                if rec["Date"] is None
                else rec["Date"].replace(".??", "").replace("????.", "")
            )

            try:
                ply = rec["PlyCount"]
                length = str(int(ply) // 2) if ply else ""
            except ValueError:
                length = ""
            eco = rec["ECO"]
            tc = rec["TimeControl"]
            variant = rec["Variant"]
            variant = variants[variant].cecp_name.capitalize() if variant else ""
            fen = rec["FEN"]

            add(
                [
                    game_id,
                    wname,
                    welo,
                    bname,
                    belo,
                    result,
                    date,
                    event,
                    site,
                    round_,
                    length,
                    eco,
                    tc,
                    variant,
                    fen,
                ]
            )

            ply = plys.get(offs) if offs in plys else 0
            self.records.append((rec, ply))

        self.set_cursor(0)

    def get_record(self, path):
        if path is None:
            return None, None
        else:
            return self.records[self.modelsort.convert_path_to_child_path(path)[0]]

    def row_activated(self, widget, path, col):
        rec, ply = self.get_record(path)
        if rec is None:
            return

        # Enable unfinished games to continue from newgamedialog
        if rec["Result"] not in UNDOABLE_STATES:
            newGameDialog.EnterNotationExtension.run()
            model = self.persp.chessfile.loadToModel(rec)
            text = pgn.save(StringIO(), model)
            newGameDialog.EnterNotationExtension.sourcebuffer.set_text(text)
            return

        self.gamemodel = GameModel()

        variant = rec["Variant"]
        if variant:
            self.gamemodel.tags["Variant"] = variant

        # Lichess exports study .pgn without White and Black tags
        wp = "" if rec["White"] is None else rec["White"]
        bp = "" if rec["Black"] is None else rec["Black"]
        p0 = (LOCAL, Human, (WHITE, wp), wp)
        p1 = (LOCAL, Human, (BLACK, bp), bp)
        self.persp.chessfile.loadToModel(rec, -1, self.gamemodel)

        self.gamemodel.endstatus = (
            self.gamemodel.status if self.gamemodel.status in UNDOABLE_STATES else None
        )
        self.gamemodel.status = WAITING_TO_START

        perspective_manager.activate_perspective("games")
        perspective = perspective_manager.get_perspective("games")
        asyncio.create_task(perspective.generalStart(self.gamemodel, p0, p1))



from pychess.Utils.const import ACTIONS


def cmp(x, y):
    return (x > y) - (x < y)


class Offer:
    def __init__(self, type_, param=None, index=None):
        assert type_ in ACTIONS, "Offer.__init__(): type not in ACTIONS: %s" % repr(
            type_
        )
        assert index is None or isinstance(
            index, int
        ), "Offer.__init__(): index not int: %s" % repr(index)
        self.type = type_
        self.param = param
        self.index = index  # for IC games

    def __hash__(self):
        return hash((self.type, self.param, self.index))

    def __cmp__(self, other):
        assert isinstance(
            other, type(self)
        ), "Offer.__cmp__(): not of type Offer: %s" % repr(other)
        return cmp(hash(self), hash(other))

    def __repr__(self):
        text = 'type="%s"' % self.type
        if self.param is not None:
            text += ", param=%s" % str(self.param)
        if self.index is not None:
            text += ", index=%s" % str(self.index)
        return "Offer(" + text + ")"



import re

from gi.repository import GObject

from pychess.Utils.const import WHITE, FISCHERRANDOMCHESS, UNSUPPORTED
from pychess.ic import (
    BLKCMD_ASSESS,
    VariantGameType,
    DEVIATION,
    GAME_TYPES,
    parse_title_hex,
    BLKCMD_UNSEEK,
    BLKCMD_SEEK,
    type_to_display_text,
    Variants,
    RATING_TYPES,
)
from pychess.ic.FICSObjects import FICSSeek
from pychess.System.Log import log

rated = "(rated|unrated)"
colors = r"(?:\[(white|black)\]\s?)?"
ratings = r"([\d\+\- ]{1,4})"
titleslist = r"(?:GM|IM|FM|WGM|WIM|WFM|TM|SR|TD|CA|C|U|D|B|T|\*)"
titleslist_re = re.compile(titleslist)
titles = r"((?:\(%s\))+)?" % titleslist
names = "([a-zA-Z]+)%s" % titles
mf = r"(?:([mf]{1,2})\s?)?"
whomatch = r"(?:(?:([-0-9+]{1,4})([\^~:\#. &])%s))" % names
whomatch_re = re.compile(whomatch)
rating_re = re.compile("[0-9]{2,}")
type_re = "(Lightning|Blitz|Standard|Suicide|Wild|Crazyhouse|Bughouse|Losers|Atomic)"


class SeekManager(GObject.GObject):
    __gsignals__ = {
        "addSeek": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "removeSeek": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "clearSeeks": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "our_seeks_removed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "seek_updated": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "assessReceived": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_line(self.on_seek_clear, "<sc>")
        self.connection.expect_line(self.on_seek_add, "<s(?:n?)> (.+)")
        self.connection.expect_line(self.on_seek_remove, r"<sr> ([\d ]+)")
        self.connection.expect_n_lines(
            self.on_our_seeks_removed,
            r"<sr> ([\d ]+)",
            r"Your seeks have been removed\.",
        )
        self.connection.expect_n_lines(
            self.on_seek_updated,
            r"Updating seek ad (\d+)(?:;?) (.*)\.",
            "",
            r"<sr> (\d+)",
            "",
            "<sn> (.+)",
        )
        self.connection.expect_n_lines(
            self.on_seek_updated,
            r"Updating seek ad (\d+)(?:;?) (.*)\.",
            r"Updating seek ad \d+(?:;?) (.*)\.",
            "",
            r"<sr> (\d+)",
            "",
            "<sn> (.+)",
        )

        self.connection.expect_n_lines(
            self.on_assess,
            r"\s*%s\s*" % type_re,
            r"\s*(\w+)\s+(\w+)\s*",
            r"\s*(\(.+\))\s+(\(.+\))\s*",
            r"\s*Win: .+",
            r"\s*Draw: .+",
            r"\s*Loss: .+",
            r"\s*New RD: .+",
        )

        self.connection.client.run_command("iset seekinfo 1")
        self.connection.client.run_command("iset seekremove 1")
        self.connection.client.run_command("iset showownseek 1")

    def seek(
        self,
        startmin,
        incsec,
        game_type,
        rated,
        ratings=(0, 9999),
        color=None,
        manual=False,
    ):
        log.debug(
            "SeekManager.seek: %s %s %s %s %s %s %s"
            % (startmin, incsec, game_type, rated, str(ratings), color, manual)
        )
        rchar = "r" if rated else "u"
        if color is not None:
            cchar = color == WHITE and "w" or "b"
        else:
            cchar = ""
        manual = "m" if manual else ""
        s = "seek %d %d %s %s %s" % (startmin, incsec, rchar, cchar, manual)
        if isinstance(game_type, VariantGameType):
            s += " " + game_type.seek_text
        if not self.connection.FatICS:
            s += " %d-%d" % (ratings[0], ratings[1])

        self.connection.client.run_command(s, show_reply=True)

    def assess(self, player1, player2, type):
        self.connection.client.run_command(f"assess {player1} {player2} /{type}")

    def on_assess(self, matchlist):
        assess = {}
        assess["type"] = matchlist[0].groups()[0]
        assess["names"] = matchlist[1].groups()[0], matchlist[1].groups()[1]
        assess["oldRD"] = matchlist[2].groups()[0], matchlist[2].groups()[1]
        assess["win"] = matchlist[3].string.split()[1:]
        assess["draw"] = matchlist[4].string.split()[1:]
        assess["loss"] = matchlist[5].string.split()[1:]
        assess["newRD"] = matchlist[6].string.split()[2:]
        self.emit("assessReceived", assess)

    on_assess.BLKCMD = BLKCMD_ASSESS

    def on_seek_add(self, match):
        # The <s> message looks like:
        # <s> index w=name_from ti=titles rt=rating t=time i=increment
        #     r=rated('r')/unrated('u') tp=type("wild/fr", "wild/4","blitz")
        #     c=color rr=rating_range(lower-upper) a=automatic?('t'/'f')
        #     f=formula_checked('t'/f')
        parts = match.groups()[0].split()
        seek = {}
        for key, value in [p.split("=") for p in parts[1:] if p]:
            seek[key] = value

        try:
            index = int(parts[0])
            player = self.connection.players.get(seek["w"])
            player.titles |= parse_title_hex(seek["ti"])
            rated = seek["r"] == "r"
            minutes = int(seek["t"])
            increment = int(seek["i"])
            rmin, rmax = (int(r) for r in seek["rr"].split("-"))
            rating = seek["rt"]
            if rating[-1] in (" ", "P", "E"):
                deviation = DEVIATION[rating[-1]]
                rating = rating[:-1]
            rating = int(rating)
            deviation = None
            automatic = seek["a"] == "t"
            color = None
            if seek["c"] == "W":
                color = "white"
            elif seek["c"] == "B":
                color = "black"
        except KeyError as e:
            log.warning(f"on_seek_add: KeyError: {repr(e)} {repr(seek)}")
            return

        try:
            gametype = GAME_TYPES[seek["tp"]]
        except KeyError:
            if self.connection.FatICS and seek["tp"] == "chess":
                # TODO: remove when fixed in FatICS
                expected_time = minutes + increment * 2 / 3
                if expected_time == 0:
                    gametype = "untimed"
                elif expected_time < 3:
                    gametype = "lightning"
                elif expected_time < 15:
                    gametype = "blitz"
                else:
                    gametype = "standard"
                gametype = GAME_TYPES[gametype]
            else:
                return
        if gametype.variant_type in UNSUPPORTED:
            return

        if (
            gametype.rating_type in RATING_TYPES
            and player.ratings[gametype.rating_type] != rating
        ):
            player.ratings[gametype.rating_type] = rating
            player.deviations[gametype.rating_type] = deviation
            player.emit("ratings_changed", gametype.rating_type, player)

        seek = FICSSeek(
            index,
            player,
            minutes,
            increment,
            rated,
            color,
            gametype,
            rmin=rmin,
            rmax=rmax,
            automatic=automatic,
        )
        self.emit("addSeek", seek)

    on_seek_add.BLKCMD = BLKCMD_SEEK

    def on_seek_clear(self, *args):
        self.emit("clearSeeks")

    def on_seek_remove(self, match):
        for key in match.groups()[0].split():
            if not key:
                continue
            self.emit("removeSeek", int(key))

    on_seek_remove.BLKCMD = BLKCMD_UNSEEK

    def on_our_seeks_removed(self, matchlist):
        self.on_seek_remove(matchlist[0])
        self.emit("our_seeks_removed")

    on_our_seeks_removed.BLKCMD = BLKCMD_UNSEEK

    def on_seek_updated(self, matchlist):
        text = matchlist[0].groups()[1]
        i = 0
        if "Updating seek ad" in matchlist[1].string:
            text += "; " + matchlist[1].groups()[0]
            i = 1
        self.on_seek_remove(matchlist[i + 2])
        self.on_seek_add(matchlist[i + 4])
        self.emit("seek_updated", text)

    on_seek_updated.BLKCMD = BLKCMD_SEEK

    def refresh_seeks(self):
        self.connection.client.run_command("iset seekinfo 1")


if __name__ == "__main__":
    assert (
        type_to_display_text("Loaded from eco/a00")
        == type_to_display_text("eco/a00")
        == "Eco A00"
    )
    assert type_to_display_text("wild/fr") == Variants.variants[FISCHERRANDOMCHESS].name
    assert type_to_display_text("blitz") == GAME_TYPES["blitz"].display_text



import subprocess
import os.path


class Command:
    def __init__(self, command, inputstr):
        self.command = command
        self.inputstr = inputstr

    def run(self, timeout=None):
        """Run a command then return: (status, output, error)."""
        status = None
        output = ""
        error = ""
        try:
            process = subprocess.Popen(
                self.command,
                universal_newlines=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                bufsize=0,
                cwd=os.path.dirname(self.command[0]),
            )
        except OSError:
            return status, output, error
        except ValueError:
            return status, output, error

        try:
            output, error = process.communicate(input=self.inputstr, timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            output, error = process.communicate()

        status = process.returncode
        return status, output, error



import sys
from operator import attrgetter
from itertools import groupby

from gi.repository import Gtk

from pychess.Utils.const import (
    WHITE,
    BLACK,
    RANDOMCHESS,
    FISCHERRANDOMCHESS,
    LOSERSCHESS,
    UNSUPPORTED,
    VARIANTS_SHUFFLE,
    VARIANTS_OTHER,
    VARIANTS_OTHER_NONSTANDARD,
)

from pychess.ic import (
    GAME_TYPES,
    VARIANT_GAME_TYPES,
    TYPE_BLITZ,
    TYPE_LIGHTNING,
    TYPE_STANDARD,
    VariantGameType,
    time_control_to_gametype,
)

from pychess.ic.FICSObjects import FICSPlayer, get_rating_range_display_text
from pychess.System import conf, uistuff
from pychess.System.prefix import addDataPrefix
from pychess.widgets import mainwindow
from pychess.widgets.ChainVBox import ChainVBox
from pychess.Variants import variants


RATING_SLIDER_STEP = 25


class SeekChallengeSection:
    seekEditorWidgets = (
        "untimedCheck",
        "minutesSpin",
        "gainSpin",
        "strengthCheck",
        "chainAlignment",
        "ratingCenterSlider",
        "toleranceSlider",
        "toleranceHBox",
        "nocolorRadio",
        "whitecolorRadio",
        "blackcolorRadio",
        # variantCombo has to come before other variant widgets so that
        # when the widget is loaded, variantRadio isn't selected by the callback,
        # overwriting the user's saved value for the variant radio buttons
        "variantCombo",
        "noVariantRadio",
        "variantRadio",
        "ratedGameCheck",
        "manualAcceptCheck",
    )

    seekEditorWidgetDefaults = {
        "untimedCheck": [False, False, False],
        "minutesSpin": [15, 10, 2],
        "gainSpin": [10, 0, 12],
        "strengthCheck": [True, True, True],
        "chainAlignment": [True, True, True],
        "ratingCenterSlider": [40, 40, 40],
        "toleranceSlider": [8, 8, 8],
        "toleranceHBox": [False, False, False],
        "variantCombo": [RANDOMCHESS, FISCHERRANDOMCHESS, LOSERSCHESS],
        "noVariantRadio": [True, True, True],
        "variantRadio": [False, False, False],
        "nocolorRadio": [True, True, True],
        "whitecolorRadio": [False, False, False],
        "blackcolorRadio": [False, False, False],
        "ratedGameCheck": [False, False, False],
        "manualAcceptCheck": [False, False, False],
    }

    seekEditorWidgetGettersSetters = {}

    def __init__(self, lounge):
        self.lounge = lounge
        self.widgets = lounge.widgets
        self.connection = lounge.connection

        self.widgets["editSeekDialog"].set_transient_for(mainwindow())
        self.widgets["challengeDialog"].set_transient_for(mainwindow())

        self.finger = None
        conf.set("numberOfFingers", 0)
        self.connection.fm.connect("fingeringFinished", self.onFinger)
        self.connection.fm.finger(self.connection.getUsername())

        self.widgets["untimedCheck"].connect("toggled", self.onUntimedCheckToggled)
        self.widgets["minutesSpin"].connect("value-changed", self.onTimeSpinChanged)
        self.widgets["gainSpin"].connect("value-changed", self.onTimeSpinChanged)
        self.onTimeSpinChanged(self.widgets["minutesSpin"])

        self.widgets["nocolorRadio"].connect("toggled", self.onColorRadioChanged)
        self.widgets["whitecolorRadio"].connect("toggled", self.onColorRadioChanged)
        self.widgets["blackcolorRadio"].connect("toggled", self.onColorRadioChanged)
        self.onColorRadioChanged(self.widgets["nocolorRadio"])

        self.widgets["noVariantRadio"].connect("toggled", self.onVariantRadioChanged)
        self.widgets["variantRadio"].connect("toggled", self.onVariantRadioChanged)
        variantcombo = self.widgets["variantCombo"]
        variantcombo.set_name("variantcombo")
        variantComboGetter, variantComboSetter = self.__initVariantCombo(variantcombo)
        self.seekEditorWidgetGettersSetters["variantCombo"] = (
            variantComboGetter,
            variantComboSetter,
        )
        self.widgets["variantCombo"].connect("changed", self.onVariantComboChanged)

        self.widgets["editSeekDialog"].connect("delete_event", lambda *a: True)
        #        self.widgets["challengeDialog"].connect("delete_event", lambda *a: True)

        self.widgets["strengthCheck"].connect("toggled", self.onStrengthCheckToggled)
        self.onStrengthCheckToggled(self.widgets["strengthCheck"])
        self.widgets["ratingCenterSlider"].connect(
            "value-changed", self.onRatingCenterSliderChanged
        )
        self.onRatingCenterSliderChanged(self.widgets["ratingCenterSlider"])
        self.widgets["toleranceSlider"].connect(
            "value-changed", self.onToleranceSliderChanged
        )
        self.onToleranceSliderChanged(self.widgets["toleranceSlider"])
        self.widgets["toleranceButton"].connect(
            "clicked", self.onToleranceButtonClicked
        )
        self.widgets["toleranceButton"].connect(
            "activate-link", lambda link_button: True
        )

        def intGetter(widget):
            return int(widget.get_value())

        self.seekEditorWidgetGettersSetters["minutesSpin"] = (intGetter, None)
        self.seekEditorWidgetGettersSetters["gainSpin"] = (intGetter, None)
        self.seekEditorWidgetGettersSetters["ratingCenterSlider"] = (intGetter, None)
        self.seekEditorWidgetGettersSetters["toleranceSlider"] = (intGetter, None)

        def toleranceHBoxGetter(widget):
            return self.widgets["toleranceHBox"].get_property("visible")

        def toleranceHBoxSetter(widget, visible):
            assert isinstance(visible, bool)
            if visible:
                self.widgets["toleranceHBox"].show()
            else:
                self.widgets["toleranceHBox"].hide()

        self.seekEditorWidgetGettersSetters["toleranceHBox"] = (
            toleranceHBoxGetter,
            toleranceHBoxSetter,
        )

        self.chainbox = ChainVBox()
        self.widgets["chainAlignment"].add(self.chainbox)

        def chainboxGetter(widget):
            return self.chainbox.active

        def chainboxSetter(widget, is_active):
            self.chainbox.active = is_active

        self.seekEditorWidgetGettersSetters["chainAlignment"] = (
            chainboxGetter,
            chainboxSetter,
        )

        self.challengee = None
        self.in_challenge_mode = False
        self.seeknumber = 1
        self.widgets["seekButton"].connect("clicked", self.onSeekButtonClicked)
        self.widgets["seekAllButton"].connect("clicked", self.onSeekAllButtonClicked)
        self.widgets["challengeButton"].connect(
            "clicked", self.onChallengeButtonClicked
        )
        self.widgets["challengeDialog"].connect(
            "delete-event", self.onChallengeDialogResponse
        )
        self.widgets["challengeDialog"].connect(
            "response", self.onChallengeDialogResponse
        )
        self.widgets["editSeekDialog"].connect(
            "response", self.onEditSeekDialogResponse
        )

        for widget in (
            "seek1Radio",
            "seek2Radio",
            "seek3Radio",
            "challenge1Radio",
            "challenge2Radio",
            "challenge3Radio",
        ):
            uistuff.keep(self.widgets[widget], widget)

        self.lastdifference = 0
        self.loading_seek_editor = False
        self.savedSeekRadioTexts = [GAME_TYPES["blitz"].display_text] * 3

        for i in range(1, 4):
            self.__loadSeekEditor(i)
            self.__writeSavedSeeks(i)
            self.widgets["seek%sRadioConfigButton" % i].connect(
                "clicked", self.onSeekRadioConfigButtonClicked, i
            )
            self.widgets["challenge%sRadioConfigButton" % i].connect(
                "clicked", self.onChallengeRadioConfigButtonClicked, i
            )

        if not self.connection.isRegistred():
            self.chainbox.active = False
            self.widgets["chainAlignment"].set_sensitive(False)
            self.widgets["chainAlignment"].set_tooltip_text(
                _(
                    "The chain button is disabled because you are logged in as a guest. Guests \
                can't establish ratings, and the chain button's state has no effect when \
                there is no rating to which to tie \"Opponent Strength\" to"
                )
            )

    def onSeekButtonClicked(self, button):
        if self.widgets["seek3Radio"].get_active():
            self.__loadSeekEditor(3)
        elif self.widgets["seek2Radio"].get_active():
            self.__loadSeekEditor(2)
        else:
            self.__loadSeekEditor(1)

        (
            minutes,
            incr,
            gametype,
            ratingrange,
            color,
            rated,
            manual,
        ) = self.__getSeekEditorDialogValues()
        self.connection.glm.seek(
            minutes, incr, gametype, rated, ratingrange, color, manual
        )

    def onSeekAllButtonClicked(self, button):
        for i in range(1, 4):
            self.__loadSeekEditor(i)
            (
                minutes,
                incr,
                gametype,
                ratingrange,
                color,
                rated,
                manual,
            ) = self.__getSeekEditorDialogValues()
            self.connection.glm.seek(
                minutes, incr, gametype, rated, ratingrange, color, manual
            )

    def onChallengeButtonClicked(self, button, player=None):
        if player is None:
            player = self.lounge.players_tab.getSelectedPlayer()
            if player is None:
                return
        self.challengee = player
        self.in_challenge_mode = True

        for i in range(1, 4):
            self.__loadSeekEditor(i)
            self.__writeSavedSeeks(i)
        self.__updateRatedGameCheck()
        if self.widgets["seek3Radio"].get_active():
            seeknumber = 3
        elif self.widgets["seek2Radio"].get_active():
            seeknumber = 2
        else:
            seeknumber = 1
        self.__updateSeekEditor(seeknumber, challengemode=True)

        self.widgets["challengeeNameLabel"].set_markup(player.getMarkup())
        self.widgets["challengeeImage"].set_from_pixbuf(player.getIcon(size=32))
        title = _("Challenge: ") + player.name
        self.widgets["challengeDialog"].set_title(title)
        self.widgets["challengeDialog"].present()

    def onChallengeDialogResponse(self, dialog, response):
        self.widgets["challengeDialog"].hide()
        if response != 5:
            return True

        if self.widgets["challenge3Radio"].get_active():
            self.__loadSeekEditor(3)
        elif self.widgets["challenge2Radio"].get_active():
            self.__loadSeekEditor(2)
        else:
            self.__loadSeekEditor(1)
        (
            minutes,
            incr,
            gametype,
            ratingrange,
            color,
            rated,
            manual,
        ) = self.__getSeekEditorDialogValues()
        self.connection.om.challenge(
            self.challengee.name, gametype, minutes, incr, rated, color
        )

    def onSeekRadioConfigButtonClicked(self, configimage, seeknumber):
        self.__showSeekEditor(seeknumber)

    def onChallengeRadioConfigButtonClicked(self, configimage, seeknumber):
        self.__showSeekEditor(seeknumber, challengemode=True)

    def onEditSeekDialogResponse(self, dialog, response):
        self.widgets["editSeekDialog"].hide()
        if response != Gtk.ResponseType.OK:
            return
        self.__saveSeekEditor(self.seeknumber)
        self.__writeSavedSeeks(self.seeknumber)

    def __updateSeekEditor(self, seeknumber, challengemode=False):
        self.in_challenge_mode = challengemode
        self.seeknumber = seeknumber
        if not challengemode:
            self.widgets["strengthFrame"].set_sensitive(True)
            self.widgets["strengthFrame"].set_tooltip_text("")
            self.widgets["manualAcceptCheck"].set_sensitive(True)
            self.widgets["manualAcceptCheck"].set_tooltip_text(
                _("If set you can refuse players accepting your seek")
            )
        else:
            self.widgets["strengthFrame"].set_sensitive(False)
            self.widgets["strengthFrame"].set_tooltip_text(
                _("This option is not applicable because you're challenging a player")
            )
            self.widgets["manualAcceptCheck"].set_sensitive(False)
            self.widgets["manualAcceptCheck"].set_tooltip_text(
                _("This option is not applicable because you're challenging a player")
            )

        self.widgets["chainAlignment"].show_all()
        self.__loadSeekEditor(seeknumber)
        self.widgets["seek%dRadio" % seeknumber].set_active(True)
        self.widgets["challenge%dRadio" % seeknumber].set_active(True)

        self.__updateYourRatingHBox()
        self.__updateRatingCenterInfoBox()
        self.__updateToleranceButton()
        self.__updateRatedGameCheck()
        self.onUntimedCheckToggled(self.widgets["untimedCheck"])

        title = (
            _("Edit Seek: ") + self.widgets["seek%dRadio" % seeknumber].get_label()[:-1]
        )
        self.widgets["editSeekDialog"].set_title(title)

    def __showSeekEditor(self, seeknumber, challengemode=False):
        self.__updateSeekEditor(seeknumber, challengemode)
        self.widgets["editSeekDialog"].present()

        # ugly hack to fix https://github.com/pychess/pychess/issues/1024
        # self.widgets["editSeekDialog"].queue_draw() doesn't work
        if sys.platform == "win32":
            self.widgets["editSeekDialog"].hide()
            allocation = self.widgets["editSeekDialog"].get_allocation()
            self.widgets["editSeekDialog"].set_size_request(
                allocation.width, allocation.height
            )
            self.widgets["editSeekDialog"].show()

    # -------------------------------------------------------- Seek Editor

    def __writeSavedSeeks(self, seeknumber):
        """Writes saved seek strings for both the Seek Panel and the Challenge Panel"""
        (
            minutes,
            gain,
            gametype,
            ratingrange,
            color,
            rated,
            manual,
        ) = self.__getSeekEditorDialogValues()
        self.savedSeekRadioTexts[seeknumber - 1] = time_control_to_gametype(
            minutes, gain
        ).display_text
        self.__writeSeekRadioLabels()
        seek = {}

        if gametype == GAME_TYPES["untimed"]:
            seek["time"] = gametype.display_text
        elif gain > 0:
            seek["time"] = _("%(minutes)d min + %(gain)d sec/move") % {
                "minutes": minutes,
                "gain": gain,
            }
        else:
            seek["time"] = _("%d min") % minutes

        if isinstance(gametype, VariantGameType):
            seek["variant"] = "%s" % gametype.display_text

        rrtext = get_rating_range_display_text(ratingrange[0], ratingrange[1])
        if rrtext:
            seek["rating"] = rrtext

        if color == WHITE:
            seek["color"] = _("White")
        elif color == BLACK:
            seek["color"] = _("Black")

        if rated and gametype is not GAME_TYPES["untimed"]:
            seek["rated"] = _("Rated")

        if manual:
            seek["manual"] = _("Manual")

        seek_ = []
        challenge = []
        challengee_is_guest = self.challengee and self.challengee.isGuest()
        for key in ("time", "variant", "rating", "color", "rated", "manual"):
            if key in seek:
                seek_.append(seek[key])
                if key in ("time", "variant", "color") or (
                    key == "rated" and not challengee_is_guest
                ):
                    challenge.append(seek[key])
        seektext = ", ".join(seek_)
        challengetext = ", ".join(challenge)

        if seeknumber == 1:
            self.widgets["seek1RadioLabel"].set_text(seektext)
            self.widgets["challenge1RadioLabel"].set_text(challengetext)
        elif seeknumber == 2:
            self.widgets["seek2RadioLabel"].set_text(seektext)
            self.widgets["challenge2RadioLabel"].set_text(challengetext)
        else:
            self.widgets["seek3RadioLabel"].set_text(seektext)
            self.widgets["challenge3RadioLabel"].set_text(challengetext)

    def __loadSeekEditor(self, seeknumber):
        self.loading_seek_editor = True
        for widget in self.seekEditorWidgets:
            if widget in self.seekEditorWidgetGettersSetters:
                uistuff.loadDialogWidget(
                    self.widgets[widget],
                    widget,
                    seeknumber,
                    get_value_=self.seekEditorWidgetGettersSetters[widget][0],
                    set_value_=self.seekEditorWidgetGettersSetters[widget][1],
                    first_value=self.seekEditorWidgetDefaults[widget][seeknumber - 1],
                )
            elif widget in self.seekEditorWidgetDefaults:
                uistuff.loadDialogWidget(
                    self.widgets[widget],
                    widget,
                    seeknumber,
                    first_value=self.seekEditorWidgetDefaults[widget][seeknumber - 1],
                )
            else:
                uistuff.loadDialogWidget(self.widgets[widget], widget, seeknumber)

        self.lastdifference = conf.get("lastdifference-%d" % seeknumber)
        self.loading_seek_editor = False

    def __saveSeekEditor(self, seeknumber):
        for widget in self.seekEditorWidgets:
            if widget in self.seekEditorWidgetGettersSetters:
                uistuff.saveDialogWidget(
                    self.widgets[widget],
                    widget,
                    seeknumber,
                    get_value_=self.seekEditorWidgetGettersSetters[widget][0],
                )
            else:
                uistuff.saveDialogWidget(self.widgets[widget], widget, seeknumber)

        conf.set("lastdifference-%d" % seeknumber, self.lastdifference)

    def __getSeekEditorDialogValues(self):
        if self.widgets["untimedCheck"].get_active():
            minutes = 0
            incr = 0
        else:
            minutes = int(self.widgets["minutesSpin"].get_value())
            incr = int(self.widgets["gainSpin"].get_value())

        if self.widgets["strengthCheck"].get_active():
            ratingrange = [0, 9999]
        else:
            center = (
                int(self.widgets["ratingCenterSlider"].get_value()) * RATING_SLIDER_STEP
            )
            tolerance = (
                int(self.widgets["toleranceSlider"].get_value()) * RATING_SLIDER_STEP
            )
            minrating = center - tolerance
            minrating = minrating > 0 and minrating or 0
            maxrating = center + tolerance
            maxrating = maxrating >= 3000 and 9999 or maxrating
            ratingrange = [minrating, maxrating]

        if self.widgets["nocolorRadio"].get_active():
            color = None
        elif self.widgets["whitecolorRadio"].get_active():
            color = WHITE
        else:
            color = BLACK

        if (
            self.widgets["noVariantRadio"].get_active()
            or self.widgets["untimedCheck"].get_active()
        ):
            gametype = time_control_to_gametype(minutes, incr)
        else:
            variant_combo_getter = self.seekEditorWidgetGettersSetters["variantCombo"][
                0
            ]
            variant = variant_combo_getter(self.widgets["variantCombo"])
            gametype = VARIANT_GAME_TYPES[variant]

        rated = (
            self.widgets["ratedGameCheck"].get_active()
            and not self.widgets["untimedCheck"].get_active()
        )
        manual = self.widgets["manualAcceptCheck"].get_active()

        return minutes, incr, gametype, ratingrange, color, rated, manual

    def __writeSeekRadioLabels(self):
        gameTypes = {
            _("Untimed"): [0, 1],
            _("Standard"): [0, 1],
            _("Blitz"): [0, 1],
            _("Lightning"): [0, 1],
        }

        for i in range(3):
            gameTypes[self.savedSeekRadioTexts[i]][0] += 1
        for i in range(3):
            if gameTypes[self.savedSeekRadioTexts[i]][0] > 1:
                labelText = "%s #%d:" % (
                    self.savedSeekRadioTexts[i],
                    gameTypes[self.savedSeekRadioTexts[i]][1],
                )
                self.widgets["seek%dRadio" % (i + 1)].set_label(labelText)
                self.widgets["challenge%dRadio" % (i + 1)].set_label(labelText)
                gameTypes[self.savedSeekRadioTexts[i]][1] += 1
            else:
                self.widgets["seek%dRadio" % (i + 1)].set_label(
                    self.savedSeekRadioTexts[i] + ":"
                )
                self.widgets["challenge%dRadio" % (i + 1)].set_label(
                    self.savedSeekRadioTexts[i] + ":"
                )

    def __updateRatingRangeBox(self):
        center = (
            int(self.widgets["ratingCenterSlider"].get_value()) * RATING_SLIDER_STEP
        )
        tolerance = (
            int(self.widgets["toleranceSlider"].get_value()) * RATING_SLIDER_STEP
        )
        min_rating = center - tolerance
        min_rating = min_rating > 0 and min_rating or 0
        max_rating = center + tolerance
        max_rating = max_rating >= 3000 and 9999 or max_rating

        self.widgets["ratingRangeMinLabel"].set_label("%d" % min_rating)
        self.widgets["ratingRangeMaxLabel"].set_label("%d" % max_rating)

        for widgetName, rating in (
            ("ratingRangeMinImage", min_rating),
            ("ratingRangeMaxImage", max_rating),
        ):
            pixbuf = FICSPlayer.getIconByRating(rating)
            self.widgets[widgetName].set_from_pixbuf(pixbuf)

        self.widgets["ratingRangeMinImage"].show()
        self.widgets["ratingRangeMinLabel"].show()
        self.widgets["dashLabel"].show()
        self.widgets["ratingRangeMaxImage"].show()
        self.widgets["ratingRangeMaxLabel"].show()
        if min_rating == 0:
            self.widgets["ratingRangeMinImage"].hide()
            self.widgets["ratingRangeMinLabel"].hide()
            self.widgets["dashLabel"].hide()
            self.widgets["ratingRangeMaxLabel"].set_label("%d↓" % max_rating)
        if max_rating == 9999:
            self.widgets["ratingRangeMaxImage"].hide()
            self.widgets["ratingRangeMaxLabel"].hide()
            self.widgets["dashLabel"].hide()
            self.widgets["ratingRangeMinLabel"].set_label("%d↑" % min_rating)
        if min_rating == 0 and max_rating == 9999:
            self.widgets["ratingRangeMinLabel"].set_label(_("Any strength"))
            self.widgets["ratingRangeMinLabel"].show()

    def __getGameType(self):
        if self.widgets["untimedCheck"].get_active():
            gametype = GAME_TYPES["untimed"]
        elif self.widgets["noVariantRadio"].get_active():
            minutes = int(self.widgets["minutesSpin"].get_value())
            gain = int(self.widgets["gainSpin"].get_value())
            gametype = time_control_to_gametype(minutes, gain)
        else:
            variant_combo_getter = self.seekEditorWidgetGettersSetters["variantCombo"][
                0
            ]
            variant = variant_combo_getter(self.widgets["variantCombo"])
            gametype = VARIANT_GAME_TYPES[variant]
        return gametype

    def __updateYourRatingHBox(self):
        gametype = self.__getGameType()
        self.widgets["yourRatingNameLabel"].set_label("(" + gametype.display_text + ")")
        rating = self.__getRating(gametype.rating_type)
        if rating is None:
            self.widgets["yourRatingImage"].clear()
            self.widgets["yourRatingLabel"].set_label(_("Unrated"))
            return
        pixbuf = FICSPlayer.getIconByRating(rating)
        self.widgets["yourRatingImage"].set_from_pixbuf(pixbuf)
        self.widgets["yourRatingLabel"].set_label(str(rating))

        center = (
            int(self.widgets["ratingCenterSlider"].get_value()) * RATING_SLIDER_STEP
        )
        rating = self.__clamp(rating)
        difference = rating - center
        if (
            self.loading_seek_editor is False
            and self.chainbox.active
            and difference != self.lastdifference
        ):
            newcenter = rating - self.lastdifference
            self.widgets["ratingCenterSlider"].set_value(
                newcenter // RATING_SLIDER_STEP
            )
        else:
            self.lastdifference = difference

    def __clamp(self, rating):
        assert isinstance(rating, int)
        mod = rating % RATING_SLIDER_STEP
        if mod > RATING_SLIDER_STEP // 2:
            return rating - mod + RATING_SLIDER_STEP
        else:
            return rating - mod

    def __updateRatedGameCheck(self):
        # on FICS, untimed games can't be rated, nor can games against a guest
        if not self.connection.isRegistred():
            self.widgets["ratedGameCheck"].set_active(False)
            sensitive = False
            self.widgets["ratedGameCheck"].set_tooltip_text(
                _("You can't play rated games because you are logged in as a guest")
            )
        elif self.widgets["untimedCheck"].get_active():
            sensitive = False
            self.widgets["ratedGameCheck"].set_tooltip_text(
                _('You can\'t play rated games because "Untimed" is checked, ')
                + _("and on FICS, untimed games can't be rated")
            )
        elif self.in_challenge_mode and self.challengee.isGuest():
            sensitive = False
            self.widgets["ratedGameCheck"].set_tooltip_text(
                _("This option is not available because you're challenging a guest, ")
                + _("and guests can't play rated games")
            )
        else:
            sensitive = True
            self.widgets["ratedGameCheck"].set_tooltip_text("")
        self.widgets["ratedGameCheck"].set_sensitive(sensitive)

    def __initVariantCombo(self, combo):
        model = Gtk.TreeStore(str)
        cellRenderer = Gtk.CellRendererText()
        combo.clear()
        combo.pack_start(cellRenderer, True)
        combo.add_attribute(cellRenderer, "text", 0)
        combo.set_model(model)

        groupNames = {
            VARIANTS_SHUFFLE: _("Shuffle"),
            VARIANTS_OTHER: _("Other (standard rules)"),
            VARIANTS_OTHER_NONSTANDARD: _("Other (non standard rules)"),
        }
        ficsvariants = [
            v
            for k, v in variants.items()
            if k in VARIANT_GAME_TYPES and v.variant not in UNSUPPORTED
        ]
        groups = groupby(ficsvariants, attrgetter("variant_group"))
        pathToVariant = {}
        variantToPath = {}
        for i, (id, group) in enumerate(groups):
            sel_iter = model.append(None, (groupNames[id],))
            for variant in group:
                subiter = model.append(sel_iter, (variant.name,))
                path = model.get_path(subiter)
                path = path.to_string()
                pathToVariant[path] = variant.variant
                variantToPath[variant.variant] = path

        # this stops group names (eg "Shuffle") from being displayed in
        # submenus
        def cellFunc(combo, cell, model, sel_iter, data):
            isChildNode = not model.iter_has_child(sel_iter)
            cell.set_property("sensitive", isChildNode)

        combo.set_cell_data_func(cellRenderer, cellFunc, None)

        def comboGetter(combo):
            path = model.get_path(combo.get_active_iter())
            path = path.to_string()
            return pathToVariant[path]

        def comboSetter(combo, variant):
            try:
                combo.set_active_iter(model.get_iter(variantToPath[variant]))
            except KeyError:
                print("Variant %s is not supported" % variant)

        return comboGetter, comboSetter

    def __getRating(self, gametype):
        if self.finger is None:
            return None
        try:
            rating = self.finger.getRating(type=gametype)
        except KeyError:  # the user doesn't have a rating for this game type
            rating = None
        return rating

    def onFinger(self, fm, finger):
        if not finger.getName() == self.connection.getUsername():
            return
        self.finger = finger

        numfingers = conf.get("numberOfFingers") + 1
        conf.set("numberOfFingers", numfingers)
        if conf.get("numberOfTimesLoggedInAsRegisteredUser") == 1 and numfingers == 1:
            standard = self.__getRating(TYPE_STANDARD)
            blitz = self.__getRating(TYPE_BLITZ)
            lightning = self.__getRating(TYPE_LIGHTNING)

            if standard is not None:
                self.seekEditorWidgetDefaults["ratingCenterSlider"][0] = (
                    standard // RATING_SLIDER_STEP
                )
            elif blitz is not None:
                self.seekEditorWidgetDefaults["ratingCenterSlider"][0] = (
                    blitz // RATING_SLIDER_STEP
                )
            if blitz is not None:
                self.seekEditorWidgetDefaults["ratingCenterSlider"][1] = (
                    blitz // RATING_SLIDER_STEP
                )
            if lightning is not None:
                self.seekEditorWidgetDefaults["ratingCenterSlider"][2] = (
                    lightning // RATING_SLIDER_STEP
                )
            elif blitz is not None:
                self.seekEditorWidgetDefaults["ratingCenterSlider"][2] = (
                    blitz // RATING_SLIDER_STEP
                )

            for i in range(1, 4):
                self.__loadSeekEditor(i)
                self.__updateSeekEditor(i)
                self.__saveSeekEditor(i)
                self.__writeSavedSeeks(i)

        self.__updateYourRatingHBox()

    def onTimeSpinChanged(self, spin):
        minutes = self.widgets["minutesSpin"].get_value_as_int()
        gain = self.widgets["gainSpin"].get_value_as_int()
        name = time_control_to_gametype(minutes, gain).display_text
        self.widgets["timeControlNameLabel"].set_label("%s" % name)
        self.__updateYourRatingHBox()

    def onUntimedCheckToggled(self, check):
        is_untimed_game = check.get_active()
        self.widgets["timeControlConfigVBox"].set_sensitive(not is_untimed_game)
        # on FICS, untimed games can't be rated and can't be a chess variant
        self.widgets["variantFrame"].set_sensitive(not is_untimed_game)
        if is_untimed_game:
            self.widgets["variantFrame"].set_tooltip_text(
                _('You can\'t select a variant because "Untimed" is checked, ')
                + _("and on FICS, untimed games have to be normal chess rules")
            )
        else:
            self.widgets["variantFrame"].set_tooltip_text("")
        self.__updateRatedGameCheck()  # sets sensitivity of widgets["ratedGameCheck"]
        self.__updateYourRatingHBox()

    def onStrengthCheckToggled(self, check):
        strengthsensitive = not check.get_active()
        self.widgets["strengthConfigVBox"].set_sensitive(strengthsensitive)

    def onRatingCenterSliderChanged(self, slider):
        center = (
            int(self.widgets["ratingCenterSlider"].get_value()) * RATING_SLIDER_STEP
        )
        pixbuf = FICSPlayer.getIconByRating(center)
        self.widgets["ratingCenterLabel"].set_label("%d" % (center))
        self.widgets["ratingCenterImage"].set_from_pixbuf(pixbuf)
        self.__updateRatingRangeBox()

        rating = self.__getRating(self.__getGameType().rating_type)
        if rating is None:
            return
        rating = self.__clamp(rating)
        self.lastdifference = rating - center

    def __updateRatingCenterInfoBox(self):
        if self.widgets["toleranceHBox"].get_property("visible") is True:
            self.widgets["ratingCenterInfoHBox"].show()
        else:
            self.widgets["ratingCenterInfoHBox"].hide()

    def __updateToleranceButton(self):
        if self.widgets["toleranceHBox"].get_property("visible") is True:
            self.widgets["toleranceButton"].set_property("label", _("Hide"))
        else:
            self.widgets["toleranceButton"].set_property("label", _("Change Tolerance"))

    def onToleranceButtonClicked(self, button):
        if self.widgets["toleranceHBox"].get_property("visible") is True:
            self.widgets["toleranceHBox"].hide()
        else:
            self.widgets["toleranceHBox"].show()
        self.__updateToleranceButton()
        self.__updateRatingCenterInfoBox()

    def onToleranceSliderChanged(self, slider):
        tolerance = (
            int(self.widgets["toleranceSlider"].get_value()) * RATING_SLIDER_STEP
        )
        self.widgets["toleranceLabel"].set_label("±%d" % tolerance)
        self.__updateRatingRangeBox()

    def onColorRadioChanged(self, radio):
        if self.widgets["nocolorRadio"].get_active():
            self.widgets["colorImage"].set_from_file(
                addDataPrefix("glade/piece-unknown.png")
            )
            self.widgets["colorImage"].set_sensitive(False)
        elif self.widgets["whitecolorRadio"].get_active():
            self.widgets["colorImage"].set_from_file(
                addDataPrefix("glade/piece-white.png")
            )
            self.widgets["colorImage"].set_sensitive(True)
        elif self.widgets["blackcolorRadio"].get_active():
            self.widgets["colorImage"].set_from_file(
                addDataPrefix("glade/piece-black.png")
            )
            self.widgets["colorImage"].set_sensitive(True)

    def onVariantRadioChanged(self, radio):
        self.__updateYourRatingHBox()

    def onVariantComboChanged(self, combo):
        self.widgets["variantRadio"].set_active(True)
        self.__updateYourRatingHBox()
        (
            min,
            gain,
            gametype,
            ratingrange,
            color,
            rated,
            manual,
        ) = self.__getSeekEditorDialogValues()
        self.widgets["variantCombo"].set_tooltip_text(
            variants[gametype.variant_type].__desc__
        )



import re

from gi.repository import GObject

from pychess.Utils.const import reprResult
from pychess.ic import (
    GAME_TYPES_BY_SHORT_FICS_NAME,
    IC_STATUS_PLAYING,
    BLKCMD_GAMES,
    GAME_TYPES,
    TITLES,
    TYPE_BLITZ,
    parse_title_hex,
    TYPE_ATOMIC,
    TYPE_WILD,
    TYPE_STANDARD,
    TYPE_LIGHTNING,
    TYPE_CRAZYHOUSE,
    TYPE_LOSERS,
    TYPE_BUGHOUSE,
    TYPE_SUICIDE,
    DEVIATION,
    STATUS,
    BLKCMD_WHO,
    IC_STATUS_NOT_AVAILABLE,
    IC_STATUS_AVAILABLE,
    parseRating,
)
from pychess.ic.FICSObjects import FICSPlayer, FICSGame
from pychess.ic.managers.BoardManager import parse_reason

rated = "(rated|unrated)"
colors = r"(?:\[(white|black)\]\s?)?"
ratings = r"([\d\+\- ]{1,4})"
titleslist = r"(?:GM|IM|FM|WGM|WIM|WFM|TM|SR|TD|CA|CM|FA|NM|C|U|D|B|T|H|\*)"
titleslist_re = re.compile(titleslist)
titles = r"((?:\(%s\))+)?" % titleslist
names = "([a-zA-Z]+)%s" % titles
mf = r"(?:([mf]{1,2})\s?)?"
whomatch = r"(?:(?:([-0-9+]{1,4})([\^~:\#. &])%s))" % names
whomatch_re = re.compile(whomatch)
whoI = (
    r"([A-Za-z]+)([\^~:\#. &])(\\d{2})" + r"(\d{1,4})([P E])" * 8 + r"(\d{1,4})([PE]?)"
)
re_whoI = re.compile(whoI)


class HelperManager(GObject.GObject):
    def __init__(self, helperconn, connection):
        GObject.GObject.__init__(self)

        self.helperconn = helperconn
        self.connection = connection

        # TODO: Examined games
        # 25 (Exam.    0 Friar          0 Friar     ) [ uu  0   0] W:  1
        # 28 ++++ TryMe       1737 Jack       [ su 30  20]  22:27 - 23:17 (29-30) W: 16
        # 2 2274 OldManII    ++++ Peshkin    [ bu  2  12]   2:34 -  1:47 (39-39) B:  3
        # 29 1622 Vman        1609 PopKid     [ sr 10  10]   1:14 -  5:10 (21-22) B: 18
        # 32 1880 Raskapov    1859 RoboDweeb  [ br  2  12]   1:04 -  1:26 ( 9-10) B: 34
        # 1 1878 Roberto     1881 baraka     [psr 45  30]  30:35 - 34:24 (22-22) W: 21
        #
        # 6 games displayed (of 23 in progress)
        self.helperconn.expect_fromto(
            self.on_game_list,
            r"(\d+) %s (\w+)\s+%s (\w+)\s+\[(p| )(%s)(u|r)\s*(\d+)\s+(\d+)\]\s*(\d:)?(\d+):(\d+)\s*-\s*(\d:)?(\d+):(\d+) \(\s*(\d+)-\s*(\d+)\) (W|B):\s*(\d+)"
            % (ratings, ratings, "|".join(GAME_TYPES_BY_SHORT_FICS_NAME.keys())),
            r"(\d+) games displayed.",
        )

        if self.helperconn.FatICS or self.helperconn.USCN:
            self.helperconn.expect_line(
                self.on_player_who, rf"{whomatch}(?:\s{{2,}}{whomatch})+"
            )
            self.helperconn.expect_line(
                self.on_player_connect, r"\[([A-Za-z]+) has connected\.\]"
            )
            self.helperconn.expect_line(
                self.on_player_disconnect, r"\[([A-Za-z]+) has disconnected\.\]"
            )
        else:
            # New ivar pin
            # http://www.freechess.org/Help/HelpFiles/new_features.html
            self.helperconn.expect_fromto(
                self.on_player_whoI, whoI, r"(\d+) Players Displayed."
            )
            self.helperconn.expect_line(self.on_player_connectI, "<wa> %s" % whoI)
            self.helperconn.expect_line(self.on_player_disconnectI, "<wd> ([A-Za-z]+)")

        self.helperconn.expect_line(
            self.on_game_add,
            r"\{Game (\d+) \(([A-Za-z]+) vs\. ([A-Za-z]+)\) (?:Creating|Continuing) (u?n?rated) ([^ ]+) match\.\}$",
        )
        self.helperconn.expect_line(
            self.on_game_remove,
            r"\{Game (\d+) \(([A-Za-z]+) vs\. ([A-Za-z]+)\) ([A-Za-z']+ .+)\} (\*|1/2-1/2|1-0|0-1)$",
        )
        self.helperconn.expect_line(
            self.on_player_unavailable, "%s is no longer available for matches." % names
        )
        self.helperconn.expect_fromto(
            self.on_player_available,
            r"%s Blitz \(%s\), Std \(%s\), Wild \(%s\), Light\(%s\), Bug\(%s\)"
            % (names, ratings, ratings, ratings, ratings, ratings),
            "is now available for matches.",
        )

        # FICS game types
        # b: blitz      l: lightning   u: untimed      e: examined game
        # s: standard   w: wild        x: atomic       z: crazyhouse
        # B: Bughouse   L: losers      S: Suicide
        if self.helperconn.FatICS or self.helperconn.USCN:
            self.helperconn.client.run_command("who")
        else:
            self.helperconn.client.run_command("who IbslwBzSLx")

        if self.helperconn.FatICS or self.helperconn.USCN:
            self.helperconn.client.run_command("games")
        else:
            self.helperconn.client.run_command("games /bslwBzSLx")

    def on_game_list(self, matchlist):
        games = []
        for match in matchlist[:-1]:
            if isinstance(match, str):
                if match:
                    parts0, parts1 = match.split("[")
                    gameno, wrating, wname, brating, bname = parts0.split()
                    private = parts1[0]
                    shorttype = parts1[1]
                    rated = parts1[2]
                    min = parts1[3:6]
                    inc = parts1[7:10]
                else:
                    continue
            else:
                (
                    gameno,
                    wrating,
                    wname,
                    brating,
                    bname,
                    private,
                    shorttype,
                    rated,
                    min,
                    inc,
                    whour,
                    wmin,
                    wsec,
                    bhour,
                    bmin,
                    bsec,
                    wmat,
                    bmat,
                    color,
                    movno,
                ) = match.groups()
            try:
                gametype = GAME_TYPES_BY_SHORT_FICS_NAME[shorttype]
            except KeyError:
                return

            wplayer = self.connection.players.get(wname)
            bplayer = self.connection.players.get(bname)
            game = FICSGame(
                wplayer,
                bplayer,
                gameno=int(gameno),
                rated=(rated == "r"),
                private=(private == "p"),
                minutes=int(min),
                inc=int(inc),
                game_type=gametype,
            )

            for player, rating in ((wplayer, wrating), (bplayer, brating)):
                if player.status != IC_STATUS_PLAYING:
                    player.status = IC_STATUS_PLAYING
                if player.game != game:
                    player.game = game
                rating = parseRating(rating)
                if player.ratings[gametype.rating_type] != rating:
                    player.ratings[gametype.rating_type] = rating
                    player.emit("ratings_changed", gametype.rating_type, player)
            game = self.connection.games.get(game, emit=False)
            games.append(game)
        self.connection.games.emit("FICSGameCreated", games)
        # print(matchlist[-1].groups()[0], len(games))

    on_game_list.BLKCMD = BLKCMD_GAMES

    def on_game_add(self, match):
        gameno, wname, bname, rated, game_type = match.groups()
        if game_type not in GAME_TYPES:
            return
        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)
        game = FICSGame(
            wplayer,
            bplayer,
            gameno=int(gameno),
            rated=(rated == "rated"),
            game_type=GAME_TYPES[game_type],
        )

        for player in (wplayer, bplayer):
            if player.status != IC_STATUS_PLAYING:
                player.status = IC_STATUS_PLAYING
            if player.game != game:
                player.game = game

        self.connection.games.get(game)

    def on_game_remove(self, match):
        gameno, wname, bname, comment, result = match.groups()
        result, reason = parse_reason(reprResult.index(result), comment, wname=wname)

        try:
            wplayer = self.connection.players.get(wname)
            wplayer.restore_previous_status()
            # no status update will be sent by
            # FICS if the player doesn't become available, so we restore
            # previous status first (not necessarily true, but the best guess)
        except KeyError:
            print("%s not in self.connections.players - creating" % wname)
            wplayer = FICSPlayer(wname)

        try:
            bplayer = self.connection.players.get(bname)
            bplayer.restore_previous_status()
        except KeyError:
            print("%s not in self.connections.players - creating" % bname)
            bplayer = FICSPlayer(bname)

        game = FICSGame(
            wplayer, bplayer, gameno=int(gameno), result=result, reason=reason
        )
        if wplayer.game is not None:
            game.rated = wplayer.game.rated
        game = self.connection.games.get(game, emit=False)

        # Our played/observed game ends are handled in main connection to prevent
        # removing them by helper connection before latest move(style12) comes from server
        if (
            game == self.connection.bm.theGameImPlaying
            or game in self.connection.bm.gamesImObserving
        ):
            return

        self.connection.games.game_ended(game)
        # Do this last to give anybody connected to the game's signals a chance
        # to disconnect from them first
        wplayer.game = None
        bplayer.game = None

    @staticmethod
    def parseTitles(titles):
        _titles = set()
        if titles:
            for title in titleslist_re.findall(titles):
                if title in TITLES:
                    _titles.add(TITLES[title])
        return _titles

    def on_player_connectI(self, match, set_online=True):
        # bslwBzSLx
        # gbtami 001411E1663P1483P1720P0P1646P0P0P1679P
        (
            name,
            status,
            titlehex,
            blitz,
            blitzdev,
            std,
            stddev,
            light,
            lightdev,
            wild,
            wilddev,
            bughouse,
            bughousedev,
            crazyhouse,
            crazyhousedev,
            suicide,
            suicidedev,
            losers,
            losersdev,
            atomic,
            atomicdev,
        ) = match.groups()
        player = self.connection.players.get(name)

        titles = parse_title_hex(titlehex)
        if not player.titles >= titles:
            player.titles |= titles

        for rating_type, elo, dev in (
            (TYPE_BLITZ, blitz, blitzdev),
            (TYPE_STANDARD, std, stddev),
            (TYPE_LIGHTNING, light, lightdev),
            (TYPE_ATOMIC, atomic, atomicdev),
            (TYPE_WILD, wild, wilddev),
            (TYPE_CRAZYHOUSE, crazyhouse, crazyhousedev),
            (TYPE_BUGHOUSE, bughouse, bughousedev),
            (TYPE_LOSERS, losers, losersdev),
            (TYPE_SUICIDE, suicide, suicidedev),
        ):
            rating = parseRating(elo)
            if player.ratings[rating_type] != rating:
                player.ratings[rating_type] = rating
                player.emit("ratings_changed", rating_type, player)
            player.deviations[rating_type] = DEVIATION[dev]

        # do last so rating info is there when notifications are generated
        status = STATUS[status]
        if player.status != status:
            player.status = status
        if set_online and not player.online:
            player.online = True

        return player

    def on_player_disconnectI(self, match):
        name = match.groups()[0]
        self.connection.players.player_disconnected(name)

    def on_player_whoI(self, matchlist):
        players = []
        for match in matchlist[:-1]:
            if isinstance(match, str):
                if match:
                    players.append(self.on_player_connectI(re_whoI.match(match)))
                else:
                    continue
            else:
                players.append(self.on_player_connectI(match))

        self.connection.players.emit("FICSPlayerEntered", players)
        # print(matchlist[-1].groups()[0], len(players))

    on_player_whoI.BLKCMD = BLKCMD_WHO

    def on_player_who(self, match):
        for blitz, status, name, titles in whomatch_re.findall(match.string):
            player = self.connection.players.get(name)
            if not player.online:
                player.online = True
            status = STATUS[status]
            if player.status != status:
                player.status = status
            titles = self.parseTitles(titles)
            if not player.titles >= titles:
                player.titles |= titles
            blitz = parseRating(blitz)
            if player.ratings[TYPE_BLITZ] != blitz:
                player.ratings[TYPE_BLITZ] = blitz
                player.emit("ratings_changed", TYPE_BLITZ, player)

    def on_player_connect(self, match):
        name = match.groups()[0]
        player = self.connection.players.get(name)
        player.online = True

    def on_player_disconnect(self, match):
        name = match.groups()[0]
        self.connection.players.player_disconnected(name)

    def on_player_unavailable(self, match):
        name, titles = match.groups()
        player = self.connection.players.get(name)
        titles = self.parseTitles(titles)
        if not player.titles >= titles:
            player.titles |= titles
        # we get here after players start a game, so we make sure that we don't
        # overwrite IC_STATUS_PLAYING
        if player.game is None and player.status not in (
            IC_STATUS_PLAYING,
            IC_STATUS_NOT_AVAILABLE,
        ):
            player.status = IC_STATUS_NOT_AVAILABLE

    def on_player_available(self, matches):
        name, titles, blitz, std, wild, light, bughouse = matches[0].groups()
        player = self.connection.players.get(name)

        if not player.online:
            player.online = True
        if player.status != IC_STATUS_AVAILABLE:
            player.status = IC_STATUS_AVAILABLE
        titles = self.parseTitles(titles)
        if not player.titles >= titles:
            player.titles |= titles

        for rating_type, rating in (
            (TYPE_BLITZ, blitz),
            (TYPE_STANDARD, std),
            (TYPE_LIGHTNING, light),
            (TYPE_WILD, wild),
            (TYPE_BUGHOUSE, bughouse),
        ):
            rating = parseRating(rating)
            if player.ratings[rating_type] != rating:
                player.ratings[rating_type] = rating
                player.emit("ratings_changed", rating_type, player)



import asyncio
from urllib.parse import urlencode
from urllib.request import urlopen

from gi.repository import GObject


from pychess.System.Log import log
from pychess.Utils.Offer import Offer
from pychess.Utils.const import ARTIFICIAL, CHAT_ACTION

from .Player import Player


class Engine(Player):
    __type__ = ARTIFICIAL
    """ "analyze" signal emits list of analysis lines. Lines are 5 element tuples.
        The first element is game ply. Second is pv string of moves. Third is a score
        relative to the engine. If no score is known, the value can be None,
        but not 0, which is a draw. Fourth is the depth of the search. Fifth is the
        nodes per second. """
    __gsignals__ = {"analyze": (GObject.SignalFlags.RUN_FIRST, None, (object,))}

    def __init__(self, md5=None):
        Player.__init__(self)
        self.md5 = md5
        self.currentAnalysis = []
        self.analyze_cid = self.connect("analyze", self.on_analysis)

    def on_analysis(self, engine, analysis):
        self.currentAnalysis = analysis

    # Offer handling

    def offer(self, offer):
        raise NotImplementedError

    def offerDeclined(self, offer):
        pass  # Ignore

    def offerWithdrawn(self, offer):
        pass  # Ignore

    def offerError(self, offer, error):
        pass  # Ignore

    # General Engine Options

    def setOptionAnalyzing(self, mode):
        self.mode = mode

    def setOptionInitialBoard(self, model):
        """If the game starts at a board other than FEN_START, it should be
        sent here. We sends a gamemodel, so the engine can load the entire
        list of moves, if any"""
        pass  # Optional

    def setOptionVariant(self, variant):
        """Inform the engine of any special variant. If the engine doesn't
        understand the variant, this will raise an error."""
        raise NotImplementedError

    def setOptionTime(self, secs, gain):
        """Seconds is the initial clock of the game.
        Gain is the amount of seconds a player gets after each move.
        If the engine doesn't support playing with time, this will fail."""
        raise NotImplementedError

    def setOptionStrength(self, strength):
        """Strength is a number [1,8] inclusive. Higher is better."""
        self.strength = strength
        raise NotImplementedError

    # Engine specific methods

    def canAnalyze(self):
        raise NotImplementedError

    def minAnalysisLines(self):
        raise NotImplementedError

    def maxAnalysisLines(self):
        raise NotImplementedError

    def requestMultiPV(self, setting):
        """Set the number of analysis lines the engine will give, if possible.

        If setting is too high, the engine's maximum will be used.
        The setting will last until the next call to requestMultiPV.
        Return value: the setting used.
        """
        raise NotImplementedError

    def getAnalysis(self):
        """Returns a list of moves, or None if there haven't yet been made an
        analysis"""
        return self.currentAnalysis

    # General chat handling

    def putMessage(self, message):
        def answer(message):
            try:
                data = (
                    urlopen(
                        "https://www.pandorabots.com/pandora/talk?botid=8d034368fe360895",
                        urlencode({"message": message, "botcust2": "x"}).encode(
                            "utf-8"
                        ),
                    )
                    .read()
                    .decode("utf-8")
                )
            except OSError as err:
                log.warning(
                    "Couldn't answer message from online bot: '%s'" % err,
                    extra={"task": self.defname},
                )
                return
            sstring = "<b>DMPGirl:</b>"
            estring = "<br>"
            answer = data[
                data.find(sstring)
                + len(sstring) : data.find(estring, data.find(sstring))
            ]
            self.emit("offer", Offer(CHAT_ACTION, answer))

        async def get_answer(message):
            loop = asyncio.get_event_loop()
            future = loop.run_in_executor(None, answer, message)
            await future

        asyncio.create_task(get_answer(message))



from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import GdkPixbuf

from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix

it = Gtk.IconTheme.get_default()


def load_icon(size, *alternatives):
    alternatives = list(alternatives)
    name = alternatives.pop(0)
    try:
        return it.load_icon(name, size, Gtk.IconLookupFlags.USE_BUILTIN)
    except Exception:
        if alternatives:
            return load_icon(size, *alternatives)
        log.warning("no %s icon in icon-theme-gnome" % name)


# Gdk.Pixbuf.new_from_file() doesn't work on Windows if path contains non ascii chars
def get_pixbuf(path, size=None):
    file = Gio.File.new_for_path(addDataPrefix(path))
    if size is None:
        return GdkPixbuf.Pixbuf.new_from_stream(file.read(None), None)
    else:
        return GdkPixbuf.Pixbuf.new_from_stream_at_scale(
            file.read(None), size, size, True, None
        )
